#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import http.client  #
import json
import threading
import sys
import time
import unittest

from dlg import utils, restutils
from dlg.manager import constants
from dlg.manager.client import MasterManagerClient
from dlg.manager.proc_daemon import DlgDaemon

_TIMEOUT = 5
IDENTITY = lambda x: x


def wait_until(
    update_condition, test_condition=IDENTITY, timeout=_TIMEOUT, interval=0.1
):
    timeout_time = time.time() + timeout
    while time.time() < timeout_time:
        output = update_condition()
        if test_condition(output):
            return output
        time.sleep(interval)
    return None


def _get_dims_from_client(client, **kwargs):
    return wait_until(lambda: client.dims(), **kwargs)


def _get_nodes_from_client(client, **kwargs):
    return wait_until(lambda: client.nodes(), **kwargs)


class TestDaemon(unittest.TestCase):
    def create_daemon(self, *args, **kwargs):
        self._daemon_t = None
        self._daemon = DlgDaemon(*args, **kwargs)

        if "noNM" not in kwargs or not kwargs["noNM"]:
            self.assertTrue(
                utils.portIsOpen(
                    "localhost", constants.NODE_DEFAULT_REST_PORT, _TIMEOUT
                ),
                "The NM did not start successfully",
            )
        if "master" in kwargs and kwargs["master"]:
            self.assertTrue(
                utils.portIsOpen(
                    "localhost", constants.MASTER_DEFAULT_REST_PORT, _TIMEOUT
                ),
                "The MM did not start successfully",
            )

        self._daemon_t = threading.Thread(
            target=lambda: self._daemon.start("localhost", 9000)
        )
        self._daemon_t.start()

        # Wait until the daemon's server has started
        # We can't simply check if the port is opened, because the server binds
        # before it is returned to us. In some tests we don't interact with it,
        # and therefore the shutdown of the daemon can occur before the server
        # is even returned to us. This would happen because portIsOpen will
        # succeed with a bound server, even if we haven't serve_forever()'d it
        # yet. In these situations shutting down the daemon will not shut down
        # the http server, and therefore the test will fail when checking that
        # the self._daemon_t is not alive anymore
        #
        # To actually avoid this we need to do some actual HTTP talk, which will
        # ensure the server is actually serving requests, and therefore already
        # in the daemon's hand
        # self.assertTrue(utils.portIsOpen('localhost', 9000, _TIMEOUT))
        try:
            restutils.RestClient("localhost", 9000, timeout=_TIMEOUT)._GET("/anything")
        except restutils.RestClientException:
            # We don't care about the result
            pass

    def tearDown(self):
        if self._daemon_t is not None:
            self._daemon.stop(_TIMEOUT)
            self._daemon_t.join(_TIMEOUT)
            self.assertFalse(
                self._daemon_t.is_alive(),
                "Daemon running thread should have finished by now",
            )
            self.assertTrue(
                utils.portIsClosed("localhost", 9000, _TIMEOUT),
                "DALiuGE Daemon REST interface should be off",
            )
        unittest.TestCase.tearDown(self)

    def test_nm_starts(self):
        # Simplest case...
        print(f"Python interpreter: {sys.executable}")
        self.create_daemon(master=False, noNM=False, disable_zeroconf=True)

    def test_mm_starts(self):
        # Start with the MM included
        self.create_daemon(master=True, noNM=False, disable_zeroconf=True)

    def test_nothing_starts(self):
        # Nothing should start now
        self.create_daemon(master=False, noNM=True, disable_zeroconf=True)
        self.assertTrue(
            utils.portIsClosed("localhost", constants.NODE_DEFAULT_REST_PORT, 0),
            "NM started but it should not have",
        )
        self.assertTrue(
            utils.portIsClosed("localhost", constants.MASTER_DEFAULT_REST_PORT, 0),
            "NM started but it should not have",
        )

    def test_zeroconf_discovery(self):

        self.create_daemon(master=True, noNM=False, disable_zeroconf=False)

        # Both managers started fine. If they zeroconf themselves correctly then
        # if we query the MM it should know about its nodes, which should have
        # one element
        mc = MasterManagerClient()
        nodes = _get_nodes_from_client(mc)
        self.assertIsNotNone(nodes)
        self.assertEqual(
            1,
            len(nodes),
            "MasterManager didn't find the NodeManager running on the same node",
        )

    def _test_zeroconf_dim_mm(self, disable_zeroconf=False):
        # Start an empty daemon, then a DIM and a Master on their own
        self.create_daemon(master=False, noNM=True, disable_zeroconf=disable_zeroconf)
        self._start("island", http.HTTPStatus.OK, {"nodes": []})
        self._start("master", http.HTTPStatus.OK)

        # Check that dim registers to MM when using zeroconf
        mc = MasterManagerClient()
        if not disable_zeroconf:

            def _test_dims(dims):
                return dims and dims["islands"]

            dims = _get_dims_from_client(mc, test_condition=_test_dims)
            self.assertIsNotNone(dims)
            return dims
        return mc.dims()

    def test_zeroconf_dim_mm(self):
        dims = self._test_zeroconf_dim_mm(disable_zeroconf=False)
        self.assertEqual(
            1,
            len(dims["islands"]),
            "MasterManager didn't find the DataIslandManager with zeroconf",
        )

    def test_without_zeroconf_dim_mm(self):
        dims = self._test_zeroconf_dim_mm(disable_zeroconf=True)
        self.assertEqual(
            0,
            len(dims["islands"]),
            "MasterManager found the DataIslandManager without zeroconf!?",
        )

    def _add_zeroconf_nm(self):
        self._start("node", http.HTTPStatus.OK)
        mc = MasterManagerClient()
        return _get_nodes_from_client(mc)

    def test_zeroconf_dim_nm_setup(self):
        """
        Sets up a mm with a node manager
        Sets up a DIM with zeroconf discovery
        Asserts that the mm attaches the nm to the discovered dim
        """
        self._test_zeroconf_dim_mm(disable_zeroconf=False)
        nodes = self._add_zeroconf_nm()
        self.assertIsNotNone(nodes)

    def test_without_zeroconf_dim_nm_setup(self):
        self._test_zeroconf_dim_mm(disable_zeroconf=True)
        nodes = self._add_zeroconf_nm()["nodes"]
        self.assertEqual(0, len(nodes))

    def test_zeroconf_nm_down(self):
        self._test_zeroconf_dim_mm(disable_zeroconf=False)
        nodes = self._add_zeroconf_nm()
        self.assertIsNotNone(nodes)
        self._stop("node", http.HTTPStatus.OK)
        mc = MasterManagerClient()

        def _test_nodes(nodes):
            return not nodes["nodes"]

        new_nodes = _get_nodes_from_client(mc, test_condition=_test_nodes)["nodes"]
        self.assertEqual(0, len(new_nodes))

    def test_start_dataisland_via_rest(self):

        self.create_daemon(master=True, noNM=False, disable_zeroconf=False)

        # Both managers started fine. If they zeroconf themselves correctly then
        # if we query the MM it should know about its nodes, which should have
        # one element
        mc = MasterManagerClient()
        nodes = _get_nodes_from_client(mc)
        self.assertIsNotNone(nodes)
        self.assertEqual(
            1,
            len(nodes),
            "MasterManager didn't find the NodeManager running on the same node",
        )

        # Check that the DataIsland starts with the given nodes
        self._start("island", http.HTTPStatus.OK, {"nodes": nodes})
        self.assertTrue(
            utils.portIsOpen("localhost", constants.ISLAND_DEFAULT_REST_PORT, _TIMEOUT),
            "The DIM did not start successfully",
        )

    def test_stop_dataisland_via_rest(self):

        # start master and island manager
        self.create_daemon(master=True, noNM=False, disable_zeroconf=False)
        mc = MasterManagerClient()
        nodes = _get_nodes_from_client(mc)
        self._start("island", http.HTTPStatus.OK, {"nodes": nodes})

        # Both managers started fine. If they zeroconf themselves correctly then
        # if we query the MM it should know about its nodes, which should have
        # one element
        mc = MasterManagerClient()
        nodes = _get_nodes_from_client(mc)
        self.assertIsNotNone(nodes)
        self.assertEqual(
            1,
            len(nodes),
            "MasterManager didn't find the NodeManager running on the same node",
        )

        # Check that the DataIsland stopped
        self._stop("island", http.HTTPStatus.OK, "")
        self.assertTrue(
            utils.portIsClosed(
                "localhost", constants.ISLAND_DEFAULT_REST_PORT, _TIMEOUT
            ),
            "The DIM did not stop successfully",
        )

    def test_stop_start_node_via_rest(self):

        # test both stop and start of NM via REST
        self.create_daemon(master=True, noNM=False, disable_zeroconf=False)

        # Both managers started fine. If they zeroconf themselves correctly then
        # if we query the MM it should know about its nodes, which should have
        # one element
        mc = MasterManagerClient()
        nodes = _get_nodes_from_client(mc)
        self.assertIsNotNone(nodes)
        self.assertEqual(
            1,
            len(nodes),
            "MasterManager didn't find the NodeManager running on the same node",
        )

        # Check that the NM stops
        self._stop("node", http.HTTPStatus.OK, "")
        self.assertTrue(
            utils.portIsClosed("localhost", constants.NODE_DEFAULT_REST_PORT, _TIMEOUT),
            "The node did not stop successfully",
        )

        # Check that the NM starts
        self._start("node", http.HTTPStatus.OK, {"pid": nodes})
        self.assertTrue(
            utils.portIsOpen("localhost", constants.NODE_DEFAULT_REST_PORT, _TIMEOUT),
            "The node did not start successfully",
        )

    def test_start_stop_master_via_rest(self):
        # test both stop and start of MASTER via REST
        self.create_daemon(master=False, noNM=False, disable_zeroconf=True)

        # Check that the MM starts
        self._start("master", http.HTTPStatus.OK)
        self.assertTrue(
            utils.portIsOpen("localhost", constants.MASTER_DEFAULT_REST_PORT, _TIMEOUT),
            "The MM did not start successfully",
        )

        # Check that the MM stops
        self._stop("master", http.HTTPStatus.OK, "")
        self.assertTrue(
            utils.portIsClosed(
                "localhost", constants.MASTER_DEFAULT_REST_PORT, _TIMEOUT
            ),
            "The MM did not stop successfully",
        )

    def test_get_dims(self):
        self.create_daemon(master=True, noNM=True, disable_zeroconf=False)
        # Check that the DataIsland starts with the given nodes
        mc = MasterManagerClient()
        dims = _get_dims_from_client(mc)
        self.assertIsNotNone(dims)
        self.assertEqual(
            0,
            len(dims["islands"]),
            "MasterManager didn't find the DataIslandManager running on the same node",
        )

    def _start(self, manager_name, expected_code, payload=None):
        conn = http.client.HTTPConnection("localhost", 9000)
        headers = {}
        if payload:
            payload = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        conn.request(
            "POST",
            f"/managers/{manager_name}/start",
            body=payload,
            headers=headers,
        )
        response = conn.getresponse()
        self.assertEqual(expected_code, response.status, response.read())
        response.close()
        conn.close()

    def _stop(self, manager_name, expected_code, payload=None):
        conn = http.client.HTTPConnection("localhost", 9000)
        headers = {}
        if payload:
            payload = json.dumps(payload)
            headers["Content-Type"] = "application/json"
        conn.request(
            "POST",
            f"/managers/{manager_name}/stop",
            body=payload,
            headers=headers,
        )
        response = conn.getresponse()
        self.assertEqual(expected_code, response.status, response.read())
        response.close()
        conn.close()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Tests how the logical blockdag construction logic works.
This refers to how parent hashes and signatures are built.

Most of these tests will be asserting the obvious, with the exception of Reproducing behaviour.
"""

import json
import unittest

import pkg_resources

from dlg.common.reproducibility.constants import ReproducibilityFlags, ALL_RMODES
from dlg.common.reproducibility.reproducibility import (
    init_lgt_repro_data,
    init_lg_repro_data,
    lg_build_blockdag,
)


def _init_graph(filename: str):
    with pkg_resources.resource_stream("test.reproducibility", filename) as file:
        lgt = json.load(file)
    return lgt


def _setup_lgt(filename: str, rmode: ReproducibilityFlags):
    lgt = _init_graph(filename)
    init_lgt_repro_data(lgt, rmode=str(rmode.value))
    init_lg_repro_data(lgt)
    leaves = lg_build_blockdag(lgt, rmode)[0]
    return lgt, leaves


class LogicalBlockdagTests(unittest.TestCase):

    def test_single(self):
        for rmode in ALL_RMODES:
            _, leaves = _setup_lgt("topoGraphs/testSingle.graph", rmode)
            self.assertTrue(len(leaves) == 1, rmode.name)

    def test_twostart(self):
        """
        A graph with two starts
        A -->
             C
        B -->
        """
        for rmode in ALL_RMODES:
            lgt, leaves = _setup_lgt("topoGraphs/testTwoStart.graph", rmode)
            parenthashes = list(
                lgt["nodeDataArray"][1]["reprodata"][rmode.name]["lg_parenthashes"].values()
            )
            if rmode == ReproducibilityFlags.REPRODUCE:
                parenthashes = list(
                    lgt["nodeDataArray"][1]["reprodata"][rmode.name][
                        "lg_parenthashes"].values()
                )
                sig0 = lgt["nodeDataArray"][0]["reprodata"][rmode.name]["lg_blockhash"]
                sig1 = lgt["nodeDataArray"][1]["reprodata"][rmode.name]["lg_blockhash"]
                sig2 = lgt["nodeDataArray"][2]["reprodata"][rmode.name]["lg_blockhash"]
                self.assertTrue(
                    len(leaves) == 1
                    and len(parenthashes) == 0
                    and sig0 == sig1
                    and sig1 == sig2
                    , rmode.name
                )
                self.assertTrue(len(parenthashes) == 0, rmode.name)
            else:
                self.assertTrue(
                    len(leaves) == 1
                    and len(parenthashes) == 2
                    and parenthashes[0] == parenthashes[1]
                    , rmode.name
                )

    def test_twoend(self):
        """
        A graph with two ends
          --> B
        A
          --> C
        """
        for rmode in ALL_RMODES:
            _, leaves = _setup_lgt("topoGraphs/testTwoEnd.graph", rmode)
            self.assertTrue(leaves[0] == leaves[1], rmode.name)

    def test_twolines(self):
        """
        A graph with two starts and two ends
        A --> B
        C --> D
        """
        for rmode in ALL_RMODES:
            _, leaves = _setup_lgt("topoGraphs/testTwoLines.graph", rmode)
            self.assertTrue(leaves[0] == leaves[1], rmode.name)

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        for rmode in ALL_RMODES:
            lgt, leaves = _setup_lgt("topoGraphs/dataFan.graph", rmode)
            if rmode != ReproducibilityFlags.REPRODUCE:
                sourcehash = lgt["nodeDataArray"][0]["reprodata"][rmode.name]["lg_blockhash"]
                parenthash1 = list(
                    lgt["nodeDataArray"][2]["reprodata"][rmode.name]["lg_parenthashes"].values()
                )
                parenthash2 = list(
                    lgt["nodeDataArray"][3]["reprodata"][rmode.name]["lg_parenthashes"].values()
                )
                self.assertTrue(parenthash1 == parenthash2 and parenthash1[0] == sourcehash,
                                rmode.name)
            else:
                sourcehash = lgt["nodeDataArray"][1]["reprodata"][rmode.name]["lg_blockhash"]
                parenthash1 = list(
                    lgt["nodeDataArray"][2]["reprodata"][rmode.name][
                        "lg_parenthashes"].values()
                )
                parenthash2 = list(
                    lgt["nodeDataArray"][3]["reprodata"][rmode.name][
                        "lg_parenthashes"].values()
                )
                self.assertTrue(parenthash1 == parenthash2 and parenthash1[0] == sourcehash)

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        for rmode in ALL_RMODES:
            lgt, _ = _setup_lgt("topoGraphs/dataFunnel.graph", rmode)
            if rmode != ReproducibilityFlags.REPRODUCE:
                sourcehash = lgt["nodeDataArray"][1]["reprodata"][rmode.name]["lg_blockhash"]
                parenthashes = list(
                    lgt["nodeDataArray"][3]["reprodata"][rmode.name]["lg_parenthashes"].values()
                )
                self.assertTrue(sourcehash == parenthashes[0] and len(parenthashes) == 1,
                                rmode.name)
            else:
                sourcehashes = [
                    lgt["nodeDataArray"][0]["reprodata"][rmode.name]["lg_blockhash"],
                    lgt["nodeDataArray"][2]["reprodata"][rmode.name]["lg_blockhash"],
                ]
                parenthashes = list(
                    lgt["nodeDataArray"][3]["reprodata"][rmode.name][
                        "lg_parenthashes"].values()
                )
                self.assertTrue(sourcehashes == parenthashes and len(parenthashes) == 2, rmode.name)

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        for rmode in ALL_RMODES:
            lgt, _ = _setup_lgt("topoGraphs/dataSandwich.graph", rmode)
            index = 0 if rmode != ReproducibilityFlags.REPRODUCE else 1
            sourcehash = lgt["nodeDataArray"][index]["reprodata"][rmode.name]["lg_blockhash"]
            parenthashes = list(
                lgt["nodeDataArray"][2]["reprodata"][rmode.name]["lg_parenthashes"].values()
            )
            self.assertTrue(sourcehash == parenthashes[0] and len(parenthashes) == 1, rmode.name)

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        for rmode in ALL_RMODES:
            lgt, _ = _setup_lgt("topoGraphs/computationSandwich.graph", rmode)
            sourcehash = lgt["nodeDataArray"][1]["reprodata"][rmode.name]["lg_blockhash"]
            parenthashes = list(
                lgt["nodeDataArray"][2]["reprodata"][rmode.name]["lg_parenthashes"].values()
            )
            if rmode != ReproducibilityFlags.REPRODUCE:
                self.assertTrue(sourcehash == parenthashes[0] and len(parenthashes) == 1,
                                rmode.name)
            else:
                self.assertEqual(0, len(parenthashes), rmode.name)


class LogicalBlockdagNothingTests(unittest.TestCase):
    """
    Tests the logical blockdag construction behaviour when asserting nothing.
    """

    rmode = ReproducibilityFlags.NOTHING

    def test_twostart(self):
        """
        A graph with two starts
        A -->
             C
        B -->
        """
        lgt, _ = _setup_lgt("topoGraphs/testTwoStart.graph", self.rmode)
        self.assertIn("reprodata", lgt["nodeDataArray"][1])

    def test_data_fan(self):
        """
        Tests that a single data source scatters its signature to downstream data drops.
        """
        lgt, _ = _setup_lgt("topoGraphs/dataFan.graph", self.rmode)
        for drop in lgt["nodeDataArray"]:
            self.assertIn("reprodata", drop)

    def test_data_funnel(self):
        """
        Tests that two data sources are collected in a single downstream data drop
        """
        lgt, _ = _setup_lgt("topoGraphs/dataFunnel.graph", self.rmode)
        for drop in lgt["nodeDataArray"]:
            self.assertIn("reprodata", drop)

    def test_data_sandwich(self):
        """
        Tests two data drops with an interim computing drop
        :return:
        """
        lgt, _ = _setup_lgt("topoGraphs/dataSandwich.graph", self.rmode)
        for drop in lgt["nodeDataArray"]:
            self.assertIn("reprodata", drop)

    def test_computation_sandwich(self):
        """
        Tests that an internal data drop surrounded by computing drops is handled correctly.
        """
        lgt, _ = _setup_lgt("topoGraphs/computationSandwich.graph", self.rmode)
        for drop in lgt["nodeDataArray"]:
            self.assertIn("reprodata", drop)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
This package contains several general-purpose data stores in form of
DROPs that we have developed as examples and for real-life use. Most of them
are based on the :class:`DataDROP`.
"""

__all__ = [
    "DirectoryContainer",
    "FileDROP",
    "InMemoryDROP",
    "SharedMemoryDROP",
    "NgasDROP",
    "RDBMSDrop",
    "PlasmaDROP",
    "PlasmaFlightDROP",
    "ParameterSetDROP",
    "EnvironmentVarDROP",
    "S3DROP",
    "DataDROP",
    "NullDROP",
]

from dlg.data.drops.directorycontainer import DirectoryContainer
from dlg.data.drops.file import FileDROP
from dlg.data.drops.memory import InMemoryDROP, SharedMemoryDROP
from dlg.data.drops.ngas import NgasDROP
from dlg.data.drops.plasma import PlasmaDROP, PlasmaFlightDROP
from dlg.data.drops.rdbms import RDBMSDrop
from dlg.data.drops.parset_drop import ParameterSetDROP
from dlg.data.drops.environmentvar_drop import EnvironmentVarDROP
from dlg.data.drops.s3_drop import S3DROP
from dlg.data.drops.data_base import DataDROP
from dlg.data.drops.data_base import NullDROP



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import os
import sqlite3
import unittest
import tempfile

from dlg.data.drops.memory import InMemoryDROP
from dlg.lifecycle.registry import RDBMSRegistry


DBFILE = tempfile.mktemp()

class TestRDBMSRegistry(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        conn = sqlite3.connect(DBFILE)  # @UndefinedVariable
        cur = conn.cursor()
        cur.execute("CREATE TABLE dlg_drop(oid varchar(64) PRIMARY KEY, phase integer)")
        cur.execute(
            "CREATE TABLE dlg_dropinstance(uid varchar(64) PRIMARY KEY, oid varchar(64), dataRef varchar(128))"
        )
        cur.execute(
            "CREATE TABLE dlg_dropaccesstime(oid varchar(64), accessTime TIMESTAMP, PRIMARY KEY (oid, accessTime))"
        )
        conn.close()

    def tearDown(self):
        os.unlink(DBFILE)

    def test_addDrop(self):
        a = InMemoryDROP("a", "a")
        registry = RDBMSRegistry("sqlite3", DBFILE)
        registry.addDrop(a)

        conn = sqlite3.connect(DBFILE)  # @UndefinedVariable
        cur = conn.cursor()
        cur.execute("SELECT oid FROM dlg_drop")
        r = cur.fetchone()
        self.assertEqual(1, len(r))
        self.assertEqual("a", r[0])
        cur.close()
        conn.close()

    def test_addDropInstances(self):

        a1 = InMemoryDROP("a", "a1")
        a2 = InMemoryDROP("a", "a2")
        registry = RDBMSRegistry("sqlite3", DBFILE)
        registry.addDrop(a1)

        uids = registry.getDropUids(a1)
        self.assertEqual(1, len(uids))
        self.assertEqual("a1", uids[0])

        registry.addDropInstance(a2)
        uids = registry.getDropUids(a1)
        uids.sort()
        self.assertEqual(2, len(uids))
        self.assertEqual("a1", uids[0])
        self.assertEqual("a2", uids[1])

        # Check accessing the database separately
        conn = sqlite3.connect(DBFILE)  # @UndefinedVariable
        cur = conn.cursor()
        cur.execute("SELECT uid FROM dlg_dropinstance WHERE oid = 'a'")
        rows = cur.fetchall()
        self.assertEqual(2, len(rows))
        uids = [r[0] for r in rows]
        uids.sort()
        self.assertEqual(2, len(uids))
        self.assertEqual("a1", uids[0])
        self.assertEqual("a2", uids[1])
        cur.close()
        conn.close()

    def test_dropAccess(self):

        a1 = InMemoryDROP("a", "a1")
        registry = RDBMSRegistry("sqlite3", DBFILE)
        registry.addDrop(a1)

        self.assertEqual(-1, registry.getLastAccess("a"))
        registry.recordNewAccess("a")

        self.assertNotEqual(-1, registry.getLastAccess("a"))



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Test the CRCApp application
"""

import os
import unittest

from dlg import droputils
from dlg.apps.crc import CRCApp, crc32
from dlg.apps.dynlib import DynlibApp
from dlg.data.drops.memory import InMemoryDROP
from dlg.data.drops.file import FileDROP

from . import test_dynlib
from .setp_up import build_shared_library
from .test_dynlib import _libname, _libpath


class CRCAppTests(unittest.TestCase):
    @unittest.skipUnless(
        build_shared_library(_libname, _libpath),
        "Example dynamic library not available",
    )
    def test_with_dynlib(self):
        """
        We test the following graph:

        A -----> B ----> C ---> D ---> E
           |        +--> F ---> G ---> H
           +------------------> I ---> J

        A and C are FileDrops; B is a DynlibApp; D, G and I are CRCApps;
        F, E, H and J are InMemoryDrops.

        The DynlibApp B copies A into C and F; therefore D, G and I should yield
        the same results, meaning that E, H and J should have the same contents.
        Similarly, A, C and F should have the same contents.

        This graph was experiencing some problems in a MacOS machine. Hopefully
        this test will shed some light on that issue and allow us to track it
        down and fix it.
        """

        # Build drops and wire them together
        a, c = (FileDROP(x, x) for x in ("a", "c"))
        b = DynlibApp("b", "b", lib=test_dynlib._libpath)
        d, g, i = (CRCApp(x, x) for x in ("d", "g", "i"))
        f, e, h, j = (InMemoryDROP(x, x) for x in ("f", "e", "h", "j"))

        for data, app in (a, b), (c, d), (f, g), (a, i):
            app.addInput(data)
        for app, data in (b, c), (b, f), (d, e), (g, h), (i, j):
            app.addOutput(data)

        # The crc32 is the same used by the CRCApp, see the imports
        data = os.urandom(1024)
        crc = str(crc32(data)).encode("utf8")

        # Execute the graph and check results
        with droputils.DROPWaiterCtx(self, (e, h, j), 5):
            a.write(data)
            a.setCompleted()

        # Data and CRCs are the expected ones
        for what, who in (data, (a, c, f)), (crc, (e, h, j)):
            for drop in who:
                self.assertEqual(what, droputils.allDropContents(drop))



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import logging
import time
import unittest

from dlg.common import dropdict, tool
from dlg.manager import client
from dlg.utils import terminate_or_kill
from test.manager import testutils

logger = logging.getLogger(__name__)
hostname = "localhost"


default_repro = {
    "rmode": "1",
    "RERUN": {
        "lg_blockhash": "x",
        "pgt_blockhash": "y",
        "pg_blockhash": "z",
    },
}
default_graph_repro = {
    "rmode": "1",
    "meta_data": {"repro_protocol": 0.1, "hashing_alg": "_sha3.sha3_256"},
    "merkleroot": "a",
    "RERUN": {
        "signature": "b",
    },
}


def add_test_reprodata(graph: list):
    for drop in graph:
        drop["reprodata"] = default_repro.copy()
    graph.append(default_graph_repro.copy())
    return graph


def memory_drop(uid):
    return dropdict(
        {
            "node": hostname,
            "oid": uid,
            "uid": uid,
            "categoryType": "Data",
            "dropclass": "dlg.data.drops.memory.InMemoryDROP",
        }
    )


def create_graph(branches, drops_per_branch):
    graph = []
    completed_uids = []
    final_apps = []
    for branch in range(branches):
        for i in range(drops_per_branch):
            data_uid = "data_%d_branch_%d" % (i, branch)
            app_uid = "app_%d_branch_%d" % (i, branch)
            data_drop = memory_drop(data_uid)
            app_drop = dropdict(
                {
                    "node": hostname,
                    "oid": app_uid,
                    "uid": app_uid,
                    "categoryType": "Application",
                    "dropclass": "dlg.apps.simple.SleepAndCopyApp",
                    "sleep_time": 0,
                }
            )
            data_drop.addConsumer(app_drop)
            graph.append(data_drop)
            graph.append(app_drop)
            if i == 0:
                completed_uids.append(data_uid)
                prev_app = data_drop
            elif i == drops_per_branch - 1:
                final_apps.append(app_drop)
            else:
                data_drop.addProducer(prev_app)

    final_drop = memory_drop("final")
    for final_app in final_apps:
        final_drop.addProducer(final_app)

    graph.append(final_drop)
    add_test_reprodata(graph)
    return graph, completed_uids


class TestBigGraph(unittest.TestCase):
    """
    A small class that simply checks that the deployment of a considerable-sized
    graph takes no longer than really expected.
    """

    def setUp(self):
        unittest.TestCase.setUp(self)

        args = ["-H", hostname, "-qq"]
        self.dmProcess = tool.start_process("nm", args)

    def tearDown(self):
        terminate_or_kill(self.dmProcess, 5)
        unittest.TestCase.tearDown(self)

    def test_submit_hugegraph(self):
        # Each branch contains a data drop and an app drop
        # All branches connect to a final data drop
        drops_per_branch = 5000
        branches = 5
        n_drops = drops_per_branch * branches * 2 + 1
        graph, completed_uids = create_graph(
            branches=branches, drops_per_branch=drops_per_branch
        )
        self.assertEqual(n_drops, len(graph) - 1)  # -1 for reprodata at end
        self._run_graph(graph, completed_uids, timeout=5)

    def _run_graph(self, graph, completed_uids, timeout=5):
        sessionId = "lala"
        restPort = 8989
        args = ["--port", str(restPort), "-N", hostname, "-qq"]

        logger.debug("Starting NM on port %d", restPort)
        c = client.NodeManagerClient(port=restPort)
        dimProcess = tool.start_process("dim", args)

        with testutils.terminating(dimProcess, timeout=timeout):
            c.create_session(sessionId)
            logger.info("Appending graph")
            c.append_graph(sessionId, graph)

            # What we are actually trying to measure with all this stuff
            start = time.time()
            c.deploy_session(sessionId, completed_uids)
            delta = time.time() - start

            # A minute is more than enough, in my PC it takes around 4 or 5 [s]
            # A minute is also way less than the ~2 [h] we observed in AWS
            self.assertLessEqual(
                delta, 60, "It took way too much time to create all drops"
            )



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import binascii
import os

import numpy as np
from pyarrow import plasma as plasma

from dlg.data.drops.data_base import DataDROP
from dlg.data.io import PlasmaIO, PlasmaFlightIO
from dlg.meta import dlg_string_param, dlg_bool_param


##
# @brief Plasma
# @details An object in a Apache Arrow Plasma in-memory object store
# @par EAGLE_START
# @param category Plasma
# @param tag daliuge
# @param plasma_path /String/ApplicationArgument/NoPort/ReadWrite//False/False/Path to the local plasma store
# @param object_id /String/ApplicationArgument/NoPort/ReadWrite//False/False/PlasmaId of the object for all compute nodes
# @param dropclass dlg.data.drops.plasma.PlasmaDROP/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param use_staging False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Enables writing to a dynamically resizeable staging buffer
# @param dummy /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Dummy port
# @par EAGLE_END
class PlasmaDROP(DataDROP):
    """
    A DROP that points to data stored in a Plasma Store
    """

    object_id: bytes = dlg_string_param("object_id", None)
    plasma_path: str = dlg_string_param("plasma_path", "/tmp/plasma")
    use_staging: bool = dlg_bool_param("use_staging", False)

    def initialize(self, **kwargs):
        super().initialize(**kwargs)
        self.plasma_path = os.path.expandvars(self.plasma_path)
        if self.object_id is None:
            self.object_id = (
                np.random.bytes(20)
                if len(self.uid) != 20
                else self.uid.encode("ascii")
            )
        elif isinstance(self.object_id, str):
            self.object_id = self.object_id.encode("ascii")

    def getIO(self):
        return PlasmaIO(
            plasma.ObjectID(self.object_id),
            self.plasma_path,
            expected_size=self._expectedSize,
            use_staging=self.use_staging,
        )

    @property
    def dataURL(self) -> str:
        return "plasma://%s" % (
            binascii.hexlify(self.object_id).decode("ascii")
        )


##
# @brief PlasmaFlight
# @details An Apache Arrow Flight server providing distributed access
# to a Plasma in-memory object store
# @par EAGLE_START
# @param category PlasmaFlight
# @param tag daliuge
# @param plasma_path /String/ApplicationArgument/NoPort/ReadWrite//False/False/Path to the local plasma store
# @param object_id /String/ApplicationArgument/NoPort/ReadWrite//False/False/PlasmaId of the object for all compute nodes
# @param dropclass dlg.data.drops.plasma.PlasmaFlightDROP/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param flight_path /String/ComponentParameter/NoPort/ReadWrite//False/False/IP and flight port of the drop owner
# @param dummy /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Dummy port
# @par EAGLE_END
class PlasmaFlightDROP(DataDROP):
    """
    A DROP that points to data stored in a Plasma Store
    """

    object_id: bytes = dlg_string_param("object_id", None)
    plasma_path: str = dlg_string_param("plasma_path", "/tmp/plasma")
    flight_path: str = dlg_string_param("flight_path", None)
    use_staging: bool = dlg_bool_param("use_staging", False)

    def initialize(self, **kwargs):
        super().initialize(**kwargs)
        self.plasma_path = os.path.expandvars(self.plasma_path)
        if self.object_id is None:
            self.object_id = (
                np.random.bytes(20)
                if len(self.uid) != 20
                else self.uid.encode("ascii")
            )
        elif isinstance(self.object_id, str):
            self.object_id = self.object_id.encode("ascii")

    def getIO(self):
        return PlasmaFlightIO(
            plasma.ObjectID(self.object_id),
            self.plasma_path,
            flight_path=self.flight_path,
            expected_size=self._expectedSize,
            use_staging=self.use_staging,
        )

    @property
    def dataURL(self) -> str:
        return "plasmaflight://%s" % (
            binascii.hexlify(self.object_id).decode("ascii")
        )



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Test the different bash-related applications
"""

import os
import shutil
import tempfile
import unittest

from dlg import droputils
from dlg.apps.bash_shell_app import (
    BashShellApp,
    StreamingInputBashApp,
    StreamingOutputBashApp,
    StreamingInputOutputBashApp,
)
from dlg.ddap_protocol import DROPStates
from dlg.data.drops.memory import InMemoryDROP
from dlg.data.drops.file import FileDROP
from dlg.droputils import DROPWaiterCtx


class BashAppTests(unittest.TestCase):
    def tearDown(self):
        shutil.rmtree("/tmp/daliuge_tfiles", True)

    def test_echo(self):
        a = FileDROP("a", "a")
        b = BashShellApp("b", "b", command="cp %i0 %o0")
        c = FileDROP("c", "c")

        b.addInput(a)
        b.addOutput(c)

        # Random data so we always check different contents
        data = os.urandom(10)
        with DROPWaiterCtx(self, c, 100):
            a.write(data)
            a.setCompleted()

        self.assertEqual(data, droputils.allDropContents(c))

        # We own the file, not root
        uid = os.getuid()
        self.assertEqual(uid, os.stat(c.path).st_uid)

    def test_quoted_commands(self):
        """
        A test to check that commands using quotes are correctly executed, which
        means that their quotes were correctly escaped when the final docker
        command was executed
        """

        def assert_message_is_correct(message, command):
            a = BashShellApp("a", "a", command=command)
            b = FileDROP("b", "b")
            a.addOutput(b)
            with DROPWaiterCtx(self, b, 100):
                a.async_execute()
            self.assertEqual(
                message.encode("utf8"), droputils.allDropContents(b)
            )

        msg = "This is a message with a single quote: '"
        assert_message_is_correct(msg, 'echo -n "{0}" > %o0'.format(msg))
        msg = 'This is a message with a double quotes: "'
        assert_message_is_correct(msg, "echo -n '{0}' > %o0".format(msg))

    def test_envvars(self):
        """Checks that the DLG_* environment variables are available to bash programs"""

        app_uid = "a"
        session_id = "session-id"

        # a fake session that has an ID
        class dummy(object):
            pass

        def assert_envvar_is_there(varname, value):
            command = "echo -n $%s > %%o0" % (varname)
            a = BashShellApp(
                app_uid, app_uid, dlg_session_id=session_id, command=command
            )
            b = FileDROP("b", "b")
            a.addOutput(b)
            with DROPWaiterCtx(self, b, 100):
                a.async_execute()
            self.assertEqual(
                value.encode("utf8"), droputils.allDropContents(b)
            )

        assert_envvar_is_there("DLG_UID", app_uid)
        assert_envvar_is_there("DLG_SESSION_ID", session_id)

    def test_reproducibility(self):
        from dlg.common.reproducibility.constants import ReproducibilityFlags
        from dlg.data.drops.data_base import NullDROP

        a = BashShellApp("a", "a", command="echo 'Hello world'")
        a.reproducibility_level = ReproducibilityFlags.RERUN
        a.setCompleted()
        b = NullDROP("b", "b")
        b.reproducibility_level = ReproducibilityFlags.RERUN
        b.setCompleted()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.REPEAT
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)

        a.reproducibility_level = ReproducibilityFlags.RECOMPUTE
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(
            a.generate_merkle_data(), {"command": "echo 'Hello world'"}
        )

        a.reproducibility_level = ReproducibilityFlags.REPRODUCE
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), {})

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_SCI
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), a.generate_rerun_data())

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_COMP
        a.commit()
        self.assertNotEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), a.generate_recompute_data())

        a.reproducibility_level = ReproducibilityFlags.REPLICATE_TOTAL
        a.commit()
        self.assertEqual(a.merkleroot, b.merkleroot)
        self.assertEqual(a.generate_merkle_data(), a.generate_repeat_data())


class StreamingBashAppTests(unittest.TestCase):
    def test_single_pipe(self):
        """
        A simple test where two bash apps are connected to each other in a
        streaming fashion. The data flows through a pipe which is created by
        the framework. The data drop in between acts only as a intermediator
        to establish the underlying communication channel.

        -------------     --------------     -------------     ----------
        | BashApp A | --> | InMemory B | --> | BashApp C | --> | File D |
        |   echo    |     | "/a/pipe"  |     |    dc     |     |        |
        -----*-------     --------------     ------*------     ----------
             |                                     |
             +-------------|named-pipe|------------+

        BashApp A writes "5 4 3 2 1" (each on a new line), which is read by cat
        and redirected to D.
        """

        output_fname = tempfile.mktemp()

        a = StreamingOutputBashApp(
            "a", "a", command=r"echo -en '5\n4\n3\n2\n1'"
        )
        b = InMemoryDROP("b", "b")
        c = StreamingInputBashApp("c", "c", command="cat > %o0")
        d = FileDROP("d", "d", filepath=output_fname)

        a.addOutput(b)
        c.addStreamingInput(b)
        c.addOutput(d)

        # Let's fire the app
        with DROPWaiterCtx(self, d, 2):
            a.async_execute()

        # The application executed, finished, and its output was recorded
        for drop in (a, b, c, d):
            self.assertEqual(
                DROPStates.COMPLETED,
                drop.status,
                "Drop %r not COMPLETED: %d" % (drop, drop.status),
            )
        self.assertEqual(
            [5, 4, 3, 2, 1],
            [int(x) for x in droputils.allDropContents(d).split(b"\n")],
        )

        # Clean up and go
        os.remove(output_fname)

    def test_two_simultaneous_pipes(self):
        """
        A more complicated test where three bash applications run at the same
        time. The first streams its output to the second one, while the second
        one streams *its* output to the third one.

        -------------     --------------     -------------     --------------     -------------     ----------
        | BashApp A | --> | InMemory B | --> | BashApp C | --> | InMemory D | --> | BashApp E | --> | File F |
        |   echo    |     | "/pipe1"   |     |    dc     |     | "/pipe2"   |     |   sort    |     |        |
        -----*-------     --------------     ----*--*-----     --------------     -----*-------     ----------
             |                                   |  |                                  |
             +-------------|named-pipe|----------+  +-----------|named-pipe|-----------+

        BashApp A writes "5 4 3 2 1" (each on a new line), which is read
        by "cat" (BashApp C). The printed results (a copy of the original) are
        streamed through D and read by "sort" (BashApp E), which writes the
        output to F.
        """

        output_fname = tempfile.mktemp()

        a = StreamingOutputBashApp(
            "a", "a", command=r"echo -en '5\n4\n3\n2\n1'"
        )
        b = InMemoryDROP("b", "b")
        c = StreamingInputOutputBashApp("c", "c", command="cat")
        d = InMemoryDROP("d", "d")
        e = StreamingInputBashApp("e", "e", command="sort -n > %o0")
        f = FileDROP("f", "f", filepath=output_fname)

        a.addOutput(b)
        b.addStreamingConsumer(c)
        c.addOutput(d)
        d.addStreamingConsumer(e)
        e.addOutput(f)

        # Let's fire the app
        with DROPWaiterCtx(self, f, 2):
            a.async_execute()

        # The application executed, finished, and its output was recorded
        for drop in (a, b, c, d, e, f):
            self.assertEqual(DROPStates.COMPLETED, drop.status)
        self.assertEqual(
            [1, 2, 3, 4, 5],
            [
                int(x)
                for x in droputils.allDropContents(f).strip().split(b"\n")
            ],
        )

        # Clean up and go
        os.remove(output_fname)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

"""
Module contains shared memory class which creates shared memory blocks for use by drops or other
DALiuGE components.
"""

import logging
import mmap
import os
import secrets
import warnings

import _posixshmem  # Does not work on Windows

LOGGER = logging.getLogger(__name__)

_O_CREX = os.O_CREAT | os.O_EXCL
_MAXNAMELENGTH = 14  # Apparently FreeBSD has this limitation


def _make_filename(name):
    if isinstance(name, str):
        name = str(name)
    if name[0] != "/":
        name = "/" + name
    if len(name) > _MAXNAMELENGTH:
        return name[:_MAXNAMELENGTH]
    return name


def _make_random_filename():
    nbytes = (_MAXNAMELENGTH - len("/")) // 2
    assert nbytes >= 2, "_SHM_NAME_PREFIX too long"
    name = "/" + secrets.token_hex(nbytes)
    assert len(name) <= _MAXNAMELENGTH
    return name


class DlgSharedMemory:
    """
    A re-implementation of multiprocessing.shared_memory for more direct usage in daliuge.
    Writes directly to shared-memory, automatically handles resizing and fetching of pre-existing
    blocks.
    If given a name, will attempt to create a new file or return a pre-existing one.
    Based heavily on Python's own shared memory implementation
    (https://github.com/python/cpython/blob/3.9/Lib/multiprocessing/shared_memory.py)
    """

    _name = None
    _fd = -1
    _mmap = None
    _buf = None
    _flags = os.O_RDWR
    _mode = 0o600

    def __init__(self, name, size=65536):
        """
        Tries to create a file with name provided. If this file exists, returns existing file.
        If name is not provided, a random name is created.
        """
        if size <= 0:
            raise ValueError("'size' must be positive")
        if name is None:
            while True:
                # Try make a random shared file
                name = _make_random_filename()
                try:
                    self._fd = _posixshmem.shm_open(
                        name, self._flags | _O_CREX, mode=self._mode
                    )
                except FileExistsError:  # Collision with other name
                    continue
                self._name = name
                break
        else:
            # Attempt to create a named file
            name = _make_filename(name)
            self._flags = _O_CREX | os.O_RDWR
            try:
                self._fd = _posixshmem.shm_open(name, self._flags, mode=self._mode)
            except FileExistsError:
                # File already exists, attempt to open in read/write mode
                self._flags = os.O_RDWR
                try:
                    self._fd = _posixshmem.shm_open(name, self._flags, mode=self._mode)
                except FileNotFoundError:
                    self.__init__(name, size)
                # Find the size of the written file
                # Needs to be set so that the file is truncated down to the correct size.
                size = os.lseek(self._fd, 0, os.SEEK_END)
                # Return to start for operations
                os.lseek(self._fd, 0, os.SEEK_SET)
            self._name = name
        try:
            os.ftruncate(self._fd, size)
            stats = os.fstat(self._fd)
            size = stats.st_size
            self._mmap = mmap.mmap(self._fd, size)
        except OSError:
            self.unlink()
            raise

        self._size = size
        self._buf = memoryview(self._mmap)

    def __del__(self):
        try:
            self.close()
        except OSError:
            pass

    def __repr__(self):
        return f"{self.__class__.__name__}({self._name!r}, size={self._size})"

    @property
    def buf(self):
        """A memoryview of contents of the shared memory block."""
        return self._buf

    @property
    def name(self):
        """Unique name that identifies the shared memory block."""
        return self._name

    @property
    def size(self):
        """Size in bytes."""
        return self._size

    def close(self):
        """
        Closes access to the shared memory but does not destroy it.
        """
        if self._buf is not None:
            self._buf.release()
            self._buf = None
        if self._mmap is not None:
            try:
                self._mmap.close()
            except BufferError:
                # WARNING: Possibly quite dodge but I cannot see a way around this.
                # If a child process creates a memory block then closes it, there will be a
                # buffer error as the parent process (child in the view of the memory) will still
                # have access.
                pass
            self._mmap = None
        if self._fd > -1:
            os.close(self._fd)
            self._fd = -1

    def unlink(self):
        """
        Requests destruction of this memory block.
        Unlink should be called once and only once across all processes
        """
        try:
            _posixshmem.shm_unlink(self._name)
        except FileNotFoundError:
            LOGGER.debug(f"{self.name} tried to unlink twice")
            warnings.warn("Cannot unlink a shared block twice", RuntimeWarning)

    def resize(self, new_size):
        """
        Replaces current file with larger block, containing old data.
        """
        old_data = bytes(self._buf)
        self.close()
        self.unlink()
        self.__init__(self._name, new_size)
        if new_size < len(old_data):
            warnings.warn("Shrinking shared block, may lose data", BytesWarning)
            self._buf[:] = old_data[0:new_size]
        else:
            self._buf[0 : len(old_data)] = old_data



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Tests the correctness of our topological sorting code.
This module does not test the final blockDAG generation.
Assumptions:
- The graphs tests are valid (not malformed)
"""

import json
import unittest

import pkg_resources

from dlg.common.reproducibility.constants import ReproducibilityFlags
from dlg.common.reproducibility.reproducibility import (
    init_lgt_repro_data,
    init_lg_repro_data,
    lg_build_blockdag,
    build_blockdag,
)

_dummydrop = {
    "oid": 1,
    "reprodata": {
        "rmode": "1",
        "RERUN": {
            "lg_blockhash": "123",
            "pgt_data": {"merkleroot": "456"},
            "pgt_parenthashes": {},
            "pgt_blockhash": "135",
            "pg_data": {"merkleroot": "bogus"},
            "pg_parenthashes": {},
            "pg_blockhash": "246",
            "rg_data": {"merkleroot": "bogus2"},
            "rg_parenthashes": {},
        }
    },
}


def _init_graph(filename):
    with pkg_resources.resource_stream("test.reproducibility", filename) as file:
        lgt = json.load(file)
    for drop in lgt["nodeDataArray"]:
        drop["reprodata"] = {}
        drop["reprodata"]["lg_parenthashes"] = []
        drop["reprodata"]["lgt_data"] = {"merkleroot": "1"}
        drop["reprodata"]["lg_data"] = {}
    return lgt


def _init_pgraph_single():
    return [_dummydrop.copy()]


def _init_pgraph_twostart():
    pgt = [_dummydrop.copy(), _dummydrop.copy(), _dummydrop.copy()]
    pgt[1]["oid"] = 2
    pgt[2]["oid"] = 3
    pgt[0]["outputs"] = [2]
    pgt[2]["outputs"] = [2]
    return pgt


def _init_pgraph_twoend():
    pgt = [_dummydrop.copy(), _dummydrop.copy(), _dummydrop.copy()]
    pgt[1]["oid"] = 2
    pgt[2]["oid"] = 3
    pgt[0]["outputs"] = [2, 3]
    return pgt


def _init_pgraph_twolines():
    pgt = [_dummydrop.copy(), _dummydrop.copy(), _dummydrop.copy(), _dummydrop.copy()]
    pgt[1]["oid"] = 2
    pgt[2]["oid"] = 3
    pgt[3]["oid"] = 4
    pgt[0]["outputs"] = [2]
    pgt[2]["outputs"] = [4]
    return pgt


class ToposortTests(unittest.TestCase):
    """
    Reads test graphs from /topoGraphs, appends some dummy reprodata then runs through the
    traversal code.
    The goal is to make sure these routines execute correct topological sorts.
    """

    def test_lg_blockdag_single(self):
        """
        Tests a single drop
        A
        """
        lgt = _init_graph("topoGraphs/testSingle.graph")
        init_lgt_repro_data(lgt, "1")
        init_lg_repro_data(lgt)
        visited = lg_build_blockdag(lgt, ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [-1])

    def test_lg_blockdag_twostart(self):
        """
        A graph with two starts
        A -->
             C
        B -->
        """
        lgt = _init_graph("topoGraphs/testTwoStart.graph")
        init_lgt_repro_data(lgt, "1")
        init_lg_repro_data(lgt)
        visited = lg_build_blockdag(lgt, ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [-3, -1, -2])

    def test_lg_blockdag_twoend(self):
        """
        A graph with two ends
          --> B
        A
          --> C
        """
        lgt = _init_graph("topoGraphs/testTwoEnd.graph")
        init_lgt_repro_data(lgt, "1")
        init_lg_repro_data(lgt)
        visited = lg_build_blockdag(lgt, ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [-1, -3, -2])

    def test_lg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        A --> B
        C --> D
        """
        lgt = _init_graph("topoGraphs/testTwoLines.graph")
        init_lgt_repro_data(lgt, "1")
        init_lg_repro_data(lgt)
        visited = lg_build_blockdag(lgt, ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [-2, -3, -1, -4])

    def test_lg_blockdag_empty(self):
        """
        Tests an empty graph. Should fail gracefully.
        """
        lgt = _init_graph("topoGraphs/testEmpty.graph")
        init_lgt_repro_data(lgt, "1")
        init_lg_repro_data(lgt)
        visited = lg_build_blockdag(lgt, ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [])

    def test_pgt_blockdag_single(self):
        """
        Tests a single drop
        1
        """
        pgt = _init_pgraph_single()
        visited = build_blockdag(pgt, "pgt", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [1])

    def test_pgt_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgt = _init_pgraph_twostart()
        visited = build_blockdag(pgt, "pgt", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [3, 1, 2])

    def test_pgt_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgt = _init_pgraph_twoend()
        visited = build_blockdag(pgt, "pgt", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [1, 3, 2])

    def test_pgt_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgt = _init_pgraph_twolines()
        visited = build_blockdag(pgt, "pgt", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [3, 4, 1, 2])

    def test_pgt_blockdag_empty(self):
        """
        Tests an empty graph. Should fail gracefully.
        """
        pgt = []
        visited = build_blockdag(pgt, "pgt", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [])

    def test_pg_blockdag_single(self):
        """
        Tests a single drop
        """
        pgr = _init_pgraph_single()
        visited = build_blockdag(pgr, "pg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [1])

    def test_pg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        pgr = _init_pgraph_twostart()
        visited = build_blockdag(pgr, "pg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [3, 1, 2])

    def test_pg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        pgr = _init_pgraph_twoend()
        visited = build_blockdag(pgr, "pg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [1, 3, 2])

    def test_pg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        pgr = _init_pgraph_twolines()
        visited = build_blockdag(pgr, "pgt", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [3, 4, 1, 2])

    def test_pg_blockdag_empty(self):
        """
        Tests an empty graph. Should fail gracefully.
        """
        pgr = []
        visited = build_blockdag(pgr, "pg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [])

    def test_rg_blockdag_single(self):
        """
        Tests a single drop
        """
        rgr = _init_pgraph_single()
        visited = build_blockdag(rgr, "rg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [1])

    def test_rg_blockdag_twostart(self):
        """
        A graph with two starts
        1 -->
             3
        2 -->
        """
        rgr = _init_pgraph_twostart()
        visited = build_blockdag(rgr, "rg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [3, 1, 2])

    def test_rg_blockdag_twoend(self):
        """
        A graph with two ends
          --> 2
        1
          --> 3
        """
        rgr = _init_pgraph_twoend()
        visited = build_blockdag(rgr, "rg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [1, 3, 2])

    def test_rg_blockdag_twolines(self):
        """
        A graph with two starts and two ends
        1 --> 2
        3 --> 4
        """
        rgr = _init_pgraph_twolines()
        visited = build_blockdag(rgr, "rg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [3, 4, 1, 2])

    def test_rg_blockdag_empty(self):
        """
        Tests an empty graph. Should fail gracefully.
        """
        rgr = []
        visited = build_blockdag(rgr, "rg", ReproducibilityFlags.RERUN)[1]
        self.assertTrue(visited == [])



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Contains a slurm client which generates slurm scripts from daliuge graphs.
"""

import datetime
import sys
import os
import subprocess
from dlg.runtime import __git_version__ as git_commit

from dlg.deploy.configs import ConfigFactory, init_tpl
from dlg.deploy.deployment_constants import DEFAULT_AWS_MON_PORT, DEFAULT_AWS_MON_HOST
from dlg.deploy.deployment_utils import find_numislands, label_job_dur


class SlurmClient:
    """
    parameters we can control:

    1. user group / account name (Required)
    2. whether to submit a graph, and if so provide graph path
    3. # of nodes (of Drop Managers)
    4. how long to run
    5. whether to produce offline graph vis
    6. whether to attach proxy for remote monitoring, and if so provide
        DLG_MON_HOST
        DLG_MON_PORT
    7. Root directory of the Log files (Required)
    """

    def __init__(
        self,
        log_root=None,
        acc=None,
        physical_graph_template_data=None,  # JSON formatted physical graph template
        logical_graph=None,
        job_dur=30,
        num_nodes=None,
        run_proxy=False,
        mon_host=DEFAULT_AWS_MON_HOST,
        mon_port=DEFAULT_AWS_MON_PORT,
        logv=1,
        facility=None,
        zerorun=False,
        max_threads=0,
        sleepncopy=False,
        num_islands=None,
        all_nics=False,
        check_with_session=False,
        submit=True,
        pip_name=None,
    ):
        self._config = ConfigFactory.create_config(facility=facility)
        self._acc = self._config.getpar("acc") if (acc is None) else acc
        self._log_root = (
            self._config.getpar("log_root") if (log_root is None) else log_root
        )
        self.modules = self._config.getpar("modules")
        self.venv = self._config.getpar("venv")
        if num_nodes is None:
            self._num_nodes = 1
        else:
            self._num_nodes = num_nodes
        self._job_dur = job_dur
        self._logical_graph = logical_graph
        self._physical_graph_template_data = physical_graph_template_data
        self._visualise_graph = False
        self._run_proxy = run_proxy
        self._mon_host = mon_host
        self._mon_port = mon_port
        self._pip_name = pip_name
        self._logv = logv
        self._zerorun = zerorun
        self._max_threads = max_threads
        self._sleepncopy = sleepncopy
        if num_islands is None:
            self._num_islands = 1
        else:
            self._num_islands = num_islands
        self._all_nics = all_nics
        self._check_with_session = check_with_session
        self._submit = submit
        self._dtstr = datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")  # .%f
        self._num_islands, self._num_nodes, self._pip_name = find_numislands(
            self._physical_graph_template_data
        )

    def get_log_dirname(self):
        """
        (pipeline name_)[Nnum_of_daliuge_nodes]_[time_stamp]
        """
        # Moved setting of dtstr to init
        # to ensure it doesn't change for this instance of SlurmClient()
        # dtstr = datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")  # .%f
        graph_name = self._pip_name.split("_")[0]  # use only the part of the graph name
        return "{0}_{1}".format(graph_name, self._dtstr)

    def create_job_desc(self, physical_graph_file):
        """
        Creates the slurm script from a physical graph
        """
        log_dir = "{0}/{1}".format(self._log_root, self.get_log_dirname())
        pardict = dict()
        pardict["VENV"] = self.venv
        pardict["NUM_NODES"] = str(self._num_nodes)
        pardict["PIP_NAME"] = self._pip_name
        pardict["SESSION_ID"] = os.path.split(log_dir)[-1]
        pardict["JOB_DURATION"] = label_job_dur(self._job_dur)
        pardict["ACCOUNT"] = self._acc
        pardict["PY_BIN"] = "python3" if pardict["VENV"] else sys.executable
        pardict["LOG_DIR"] = log_dir
        pardict["GRAPH_PAR"] = (
            '-L "{0}"'.format(self._logical_graph)
            if self._logical_graph
            else '-P "{0}"'.format(physical_graph_file)
            if physical_graph_file
            else ""
        )
        pardict["PROXY_PAR"] = (
            "-m %s -o %d" % (self._mon_host, self._mon_port) if self._run_proxy else ""
        )
        pardict["GRAPH_VIS_PAR"] = "-d" if self._visualise_graph else ""
        pardict["LOGV_PAR"] = "-v %d" % self._logv
        pardict["ZERORUN_PAR"] = "-z" if self._zerorun else ""
        pardict["MAXTHREADS_PAR"] = "-t %d" % self._max_threads
        pardict["SNC_PAR"] = "--app 1" if self._sleepncopy else "--app 0"
        pardict["NUM_ISLANDS_PAR"] = "-s %d" % self._num_islands
        pardict["ALL_NICS"] = "-u" if self._all_nics else ""
        pardict["CHECK_WITH_SESSION"] = "-S" if self._check_with_session else ""
        pardict["MODULES"] = self.modules

        job_desc = init_tpl.safe_substitute(pardict)
        return job_desc

    def submit_job(self):
        """
        Submits the slurm script to the cluster
        """
        log_dir = "{0}/{1}".format(self._log_root, self.get_log_dirname())
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        physical_graph_file_name = "{0}/{1}".format(log_dir, self._pip_name)
        with open(physical_graph_file_name, "w") as physical_graph_file:
            physical_graph_file.write(self._physical_graph_template_data)
            physical_graph_file.close()

        job_file_name = "{0}/jobsub.sh".format(log_dir)
        job_desc = self.create_job_desc(physical_graph_file_name)
        with open(job_file_name, "w") as job_file:
            job_file.write(job_desc)

        with open(os.path.join(log_dir, "git_commit.txt"), "w") as git_file:
            git_file.write(git_commit)
        if self._submit:
            os.chdir(log_dir)  # so that slurm logs will be dumped here
            print(subprocess.check_output(["sbatch", job_file_name]))
        else:
            print(f"Created job submission script {job_file_name}")



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Similar to chilesdospec, but it generates a series of DockerApp DROPs instead of
directly generating Clean, Flux and Split DROPs.
"""

import os
import sys
import time

from dlg.common import dropdict
from dlg.manager.client import DataIslandManagerClient

LOCAL_FILES = os.path.dirname(os.path.realpath(__file__))
CASAPY = "/opt/casa-release-4.4.0-el6/"
VIS_ROOT = "/opt/data/chiles/input/"
VIS_OUT = "/opt/data/chiles/output/"
CUBE_OUT = "/opt/data/chiles/output/"
CUBE_NAME = "cube1408~1412"

# Internal AWS IP addresses. ch05 does the Clean-ing
VIS = [
    (VIS_ROOT + "20131025_951_4_FINAL_PRODUCTS", VIS_OUT + "20131025_951_4/"),
    (VIS_ROOT + "20131031_951_4_FINAL_PRODUCTS", VIS_OUT + "20131031_951_4/"),
    (VIS_ROOT + "20131121_946_6_FINAL_PRODUCTS", VIS_OUT + "20131121_946_6/"),
    (VIS_ROOT + "20140105_946_6_FINAL_PRODUCTS", VIS_OUT + "20140105_946_6/"),
]


def fileDropSpec(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": str(uid),
            "categoryType": "Data",
            "dropclass": "dlg.data.drops.file.FileDROP",
            "node": "localhost",
            "island": "localhost",
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def directorySpec(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": str(uid),
            "categoryType": "container",
            "container": "dlg.drop.DirectoryContainer",
            "node": "localhost",
            "island": "localhost",
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def casapyDockerAppSpec(uid, script):
    cmd = (
        "cd; "
        + os.path.join(CASAPY, "casapy")
        + ' --colors=NoColor --nologger --nogui -c "%s"' % (script)
    )
    return dropdict(
        {
            "oid": str(uid),
            "categoryType": "Application",
            "dropclass": "dlg.apps.dockerapp.DockerApp",
            "image": "dfms/casapy_centos7_dfms:0.1",
            "command": cmd,
            "user": "dfms",
            "node": "localhost",
            "island": "localhost",
        }
    )


def fluxSpec(uid, **kwargs):
    script = "ia.open('%i0.image');"
    script += "flux = ia.pixelvalue([128,128,0,179])['value']['value'];"
    script += "f = open('%o0','w'); f.write(str(flux)); f.close()"
    return casapyDockerAppSpec(uid, script)


def splitSpec(uid, **kwargs):
    transform_args = kwargs.copy()
    transform_args.update({"vis": "%i0", "outputvis": "%o0"})
    script = "import shutil; shutil.rmtree('%o0', True); "
    script += "mstransform(**{})".format(repr(transform_args))
    return casapyDockerAppSpec(uid, script)


def cleanSpec(uid, **kwargs):
    clean_args = kwargs.copy()
    clean_args.update(
        {
            "niter": 0,
            "mask": "",
            "modelimage": "",
            "imagename": "%o0",
            "vis": ["%i0", "%i1", "%i2", "%i3"],
            "threshold": "0Jy",
        }
    )
    script = "import shutil; shutil.rmtree('%o0', True); "
    script += "clean(**{})".format(repr(clean_args))
    return casapyDockerAppSpec(uid, script)


if __name__ == "__main__":
    try:
        sessionId = "Chiles-Docker-%s" % (time.time(),)
        droplist = []

        flux_out = fileDropSpec("Flux", dirname=VIS_OUT)
        droplist.append(flux_out)
        flux = fluxSpec("FluxExtractor")
        droplist.append(flux)

        cl = cleanSpec(
            "Cleaning",
            field="deepfield",
            mode="frequency",
            restfreq="1420.405752MHz",
            nchan=-1,
            start="",
            width="",
            interpolation="nearest",
            gain=0.1,
            imsize=[256],
            cell=["1.0arcsec"],
            phasecenter="10h01m53.9,+02d24m52s",
            weighting="natural",
            usescratch=False,
        )
        droplist.append(cl)

        image_out = directorySpec(
            "CleanedImage", dirname=CUBE_OUT + CUBE_NAME, check_exists=False
        )
        droplist.append(image_out)
        cl.addOutput(image_out)
        flux.addInput(image_out)
        flux.addOutput(flux_out)

        for i, v in enumerate(VIS):
            vis_in = directorySpec("vis%d" % (i), dirname=v[0], check_exists=False)
            split_out = directorySpec(
                "SplitOutput_%d" % (i), dirname=v[1], check_exists=False
            )
            sp = splitSpec(
                "Splitting_%d" % (i),
                regridms=True,
                restfreq="1420.405752MHz",
                mode="frequency",
                nchan=256,
                outframe="lsrk",
                interpolation="linear",
                start="1408MHz",
                width="1412kHz",
                veltype="radio",
                spw="",
                combinespws=True,
                nspw=1,
                createmms=False,
                datacolumn="data",
            )
            sp.addInput(vis_in)
            sp.addOutput(split_out)
            cl.addInput(split_out)

            droplist.append(vis_in)
            droplist.append(split_out)
            droplist.append(sp)

        c = DataIslandManagerClient()
        c.create_session(sessionId)
        c.append_graph(sessionId, droplist)
        c.deploy_session(sessionId)

    except Exception as e:
        import traceback

        traceback.print_exc()
        sys.exit(-1)

    sys.exit(0)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import os
import unittest

from dlg import droputils
from dlg import utils
from dlg.apps.socket_listener import SocketListenerApp
from dlg.ddap_protocol import DROPStates
from dlg.data.drops.memory import InMemoryDROP
from dlg.droputils import DROPWaiterCtx
from test.test_drop import SumupContainerChecksum

try:
    from crc32c import crc32c  # @UnusedImport
except:
    from binascii import crc32  # @Reimport


class TestSocketListener(unittest.TestCase):
    def _test_socket_listener(self, **kwargs):
        """
        A simple test to check that SocketListenerApps are indeed working as
        expected; that is, they write the data they receive into their output,
        and finish when the connection is closed from the client side

        The data flow diagram looks like this:

        A --> B --> C --> D
        """

        host = "localhost"
        port = 9933
        data = os.urandom(1025)

        a = SocketListenerApp("oid:A", "uid:A", host=host, port=port, **kwargs)
        b = InMemoryDROP("oid:B", "uid:B")
        c = SumupContainerChecksum("oid:C", "uid:C")
        d = InMemoryDROP("oid:D", "uid:D")
        a.addOutput(b)
        b.addConsumer(c)
        c.addOutput(d)

        # Create the socket, write, and close the connection, allowing
        # A to move to COMPLETED
        with DROPWaiterCtx(self, d, 3):  # That's plenty of time
            a.async_execute()
            utils.write_to(host, port, data, 1)

        for drop in [a, b, c, d]:
            self.assertEqual(DROPStates.COMPLETED, drop.status)

        # Our expectations are fulfilled!
        bContents = droputils.allDropContents(b)
        dContents = int(droputils.allDropContents(d))
        self.assertEqual(data, bContents)
        self.assertEqual(crc32c(data, 0), dContents)

    def test_socket_listener(self):
        self._test_socket_listener()

    def test_socket_listener_integer_with_bufsize(self):
        for bufsize in (4096, "4096"):
            self._test_socket_listener(bufsize=bufsize)

    def test_invalid(self):

        # Shouldn't allow inputs
        a = SocketListenerApp("a", "a", port=64 * 1024)
        a.addOutput(InMemoryDROP("c", "c"))
        self.assertRaises(Exception, a.addInput, InMemoryDROP("b", "b"))
        self.assertRaises(Exception, a.addStreamingInput, InMemoryDROP("b", "b"))

        # Shouldn't be able to open ports > 64k - 1
        a.execute()
        self.assertEqual(a.status, DROPStates.ERROR)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Created on 22 Jun 2015

@author: rtobar
"""

import os
import shutil
import tempfile
import time
import unittest

from dlg.ddap_protocol import DROPStates, DROPPhases
from dlg.apps.app_base import BarrierAppDROP
from dlg.data.drops.directorycontainer import DirectoryContainer
from dlg.data.drops.file import FileDROP
from dlg.droputils import DROPWaiterCtx
from dlg.lifecycle import dlm


class TestDataLifecycleManager(unittest.TestCase):
    def tearDown(self):
        shutil.rmtree("/tmp/daliuge_tfiles", True)
        shutil.rmtree("/tmp/sdp-hsm", True)

    def _writeAndClose(self, drop):
        drop.write(b" ")
        # all DROPs submitted to this method have expectedSize=1, so this
        # will trigger the change to COMPLETED

    def test_basicCreation(self):
        manager = dlm.DataLifecycleManager()
        manager.startup()
        manager.cleanup()

    def test_dropAddition(self):
        with dlm.DataLifecycleManager() as manager:
            drop = FileDROP("oid:A", "uid:A1", expectedSize=10)
            manager.addDrop(drop)

    def test_dropCompleteTriggersReplication(self):
        with dlm.DataLifecycleManager(enable_drop_replication=True) as manager:
            drop = FileDROP("oid:A", "uid:A1", expectedSize=1)
            manager.addDrop(drop)
            self._writeAndClose(drop)

            # The call to close() should have turned it into a SOLID object
            # because the DLM replicated it
            self.assertEqual(DROPPhases.SOLID, drop.phase)
            self.assertEqual(2, len(manager.getDropUids(drop)))

            # Try the same with a non-persisted data object, it shouldn't be replicated
            drop = FileDROP("oid:B", "uid:B1", expectedSize=1, persist=False)
            manager.addDrop(drop)
            self._writeAndClose(drop)
            self.assertEqual(DROPPhases.GAS, drop.phase)
            self.assertEqual(1, len(manager.getDropUids(drop)))

    def test_expiringNormalDrop(self):
        with dlm.DataLifecycleManager(check_period=0.5) as manager:
            drop = FileDROP("oid:A", "uid:A1", expectedSize=1, lifespan=0.5)
            manager.addDrop(drop)

            # Writing moves the DROP to COMPLETE
            self._writeAndClose(drop)

            # Wait now, the DROP should be moved by the DLM to EXPIRED
            time.sleep(1)

            self.assertEqual(DROPStates.EXPIRED, drop.status)

    def test_lostDrop(self):
        with dlm.DataLifecycleManager(check_period=0.5) as manager:
            drop = FileDROP(
                "oid:A", "uid:A1", expectedSize=1, lifespan=10, persist=False
            )
            manager.addDrop(drop)
            self._writeAndClose(drop)

            # "externally" remove the file, its contents now don't exist
            os.unlink(drop.path)

            # Let the DLM do its work
            time.sleep(1)

            # Check that the DROP is marked as LOST
            self.assertEqual(DROPPhases.LOST, drop.phase)

    def test_cleanupExpiredDrops(self):
        with dlm.DataLifecycleManager(check_period=0.5, cleanup_period=2) as manager:
            drop = FileDROP(
                "oid:A", "uid:A1", expectedSize=1, lifespan=1, persist=False
            )
            manager.addDrop(drop)
            self._writeAndClose(drop)

            # Wait 2 seconds, the DROP is still COMPLETED
            time.sleep(0.5)
            self.assertEqual(DROPStates.COMPLETED, drop.status)
            self.assertTrue(drop.exists())

            # Wait 5 more second, now it should be expired but still there
            time.sleep(1)
            self.assertEqual(DROPStates.EXPIRED, drop.status)
            self.assertTrue(drop.exists())

            # Wait 2 more seconds, now it should have been deleted
            time.sleep(1)
            self.assertEqual(DROPStates.DELETED, drop.status)
            self.assertFalse(drop.exists())

    def test_expireAfterUse(self):
        """
        Simple test for the expireAfterUse flag. Two DROPs are created with
        different values, and after they are used we check whether their data
        is still there or not
        """

        class MyApp(BarrierAppDROP):
            def run(self):
                pass

        with dlm.DataLifecycleManager(check_period=0.5, cleanup_period=2) as manager:
            a = DirectoryContainer(
                "a",
                "a",
                persist=False,
                expireAfterUse=True,
                dirname=tempfile.mkdtemp(),
            )
            b_dirname = tempfile.mkdtemp()
            b = DirectoryContainer(
                "b",
                "b",
                persist=False,
                expireAfterUse=False,
                dirname=b_dirname,
            )
            c = MyApp("c", "c")
            d = MyApp("d", "d")
            a.addConsumer(c)
            a.addConsumer(d)
            b.addConsumer(c)
            b.addConsumer(d)

            manager.addDrop(a)
            manager.addDrop(b)
            manager.addDrop(b)
            manager.addDrop(c)

            # Make sure all consumers are done
            with DROPWaiterCtx(self, [c, d], 1):
                a.setCompleted()
                b.setCompleted()

            # Both directories should be there, but after cleanup A's shouldn't
            # be there anymore
            self.assertTrue(a.exists())
            self.assertTrue(b.exists())
            time.sleep(2.5)
            self.assertFalse(a.exists())
            self.assertTrue(b.exists())
            b.delete()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module tests shared memory manager.
"""

import sys
import unittest

if sys.version_info >= (3, 8):
    from dlg.manager.shared_memory_manager import DlgSharedMemoryManager


@unittest.skipIf(sys.version_info < (3, 8), "Shared memory does not work < python 3.8")
class TestSharedMemoryManager(unittest.TestCase):
    """
    Tests the DlgSharedMemory Manager in typical and atypical usecases
    """

    def test_register_session(self):
        """
        SMM should successfully register a session with no drops
        """
        manager = DlgSharedMemoryManager()
        manager.register_session("session1")
        self.assertTrue(len(manager.drop_names), 1)
        manager.shutdown_all()

    def test_register_drop(self):
        """
        SMM should register a drop with a pre-existing session
        """
        manager = DlgSharedMemoryManager()
        manager.register_session("session1")
        manager.register_drop("A", "session1")
        self.assertTrue(len(manager.drop_names["session1"]), 1)
        manager.shutdown_all()

    def test_register_drop_nonsession(self):
        """
        SMM should register the session for a drop if it did not already exist
        """
        manager = DlgSharedMemoryManager()
        manager.register_drop("A", "session1")
        self.assertTrue((len(manager.drop_names["session1"]), 1))
        manager.shutdown_all()

    def test_shutdown_session_existing(self):
        """
        SMM should successfully remove a session when requested
        """
        manager = DlgSharedMemoryManager()
        manager.register_session("session1")
        manager.register_session("session2")
        self.assertTrue(len(manager.drop_names.keys()), 2)
        manager.shutdown_session("session1")
        self.assertTrue(len(manager.drop_names.keys()), 1)
        manager.shutdown_all()

    def test_shutdown_session_nonexistent(self):
        """
        SMM should not error when attempting to shutdown a non-existent session
        """
        manager = DlgSharedMemoryManager()
        try:
            manager.shutdown_session("session1")
        except KeyError:
            self.fail("Manager errored when shutting down nonexistent session")

    def test_shutdown_all(self):
        """
        SMM should be able to remove all sessions and drop references when shutdown
        """
        manager = DlgSharedMemoryManager()
        manager.register_session("session1")
        manager.register_session("session2")
        self.assertEqual(len(manager.drop_names.keys()), 2)
        manager.shutdown_all()
        self.assertEqual(len(manager.drop_names.keys()), 0)

    def test_shutdown_all_nosessions(self):
        """
        SMM should successfully shutdown without error when no sessions or drops are present
        """
        manager = DlgSharedMemoryManager()
        try:
            manager.shutdown_all()
        except KeyError:
            self.fail("Manager errored when shutting down empty")



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import functools
import io
import json
import os
import tempfile
import unittest
import zlib

from dlg import utils


class TestUtils(unittest.TestCase):
    def test_zlib_uncompressed_stream(self):
        fname = tempfile.mktemp()
        with open(fname, "wb") as f:
            f.write(zlib.compress(b"abc"))

        # Read parts from the beginning
        for b, n in ((b"abc", 3), (b"ab", 2), (b"a", 1)):
            with open(fname, "rb") as f:
                s = utils.ZlibUncompressedStream(f)
                self.assertEqual(b, s.read(n))

        # Read whole contents by successive calls
        with open(fname, "rb") as f:
            s = utils.ZlibUncompressedStream(f)
            self.assertEqual(b"a", s.read(1))
            self.assertEqual(b"b", s.read(1))
            self.assertEqual(b"c", s.read(1))
            self.assertEqual(0, s.buflen)

        os.remove(fname)

        # Try now with bigger data sizes
        for size in [2**x + y for x in range(3, 18) for y in (-1, 0, 1)]:
            original_bytes = os.urandom(size)
            compressed_bytes = zlib.compress(original_bytes)

            # Try first with whole reads
            compressed_stream = io.BytesIO(compressed_bytes)
            uncompressed_stream = utils.ZlibUncompressedStream(
                compressed_stream
            )
            b = uncompressed_stream.read()
            self.assertEqual(size, len(b))
            self.assertEqual(original_bytes, b)

            # Now read little by little
            read_size = min(size // 4, 1024)
            b = b""
            compressed_stream = io.BytesIO(compressed_bytes)
            uncompressed_stream = utils.ZlibUncompressedStream(
                compressed_stream
            )
            for u in iter(
                functools.partial(uncompressed_stream.read, read_size), b""
            ):
                b += u
            self.assertEqual(size, len(b))
            self.assertEqual(original_bytes, b)

    def test_zlib_compressed_stream_writer(self):
        compressed_ref = zlib.compress(b"abcd")

        # Read parts from the beginning
        for x in range(1, len(compressed_ref)):
            bytesio = io.BytesIO(b"abcd")
            s = utils.ZlibCompressedStream(bytesio)
            compressed = s.read(x)
            self.assertEqual(x, len(compressed))
            self.assertEqual(compressed_ref[0:x], compressed[0:x])

        # Try now with bigger data sizes
        for size in [2**x + y for x in range(3, 18) for y in (-1, 0, 1)]:
            original_bytes = os.urandom(size)
            compressed_bytes = zlib.compress(original_bytes)

            # Try first with whole reads
            uncompressed_stream = io.BytesIO(original_bytes)
            compressed_stream = utils.ZlibCompressedStream(uncompressed_stream)
            b = compressed_stream.read()
            self.assertEqual(
                len(compressed_bytes),
                len(b),
                "Incorrect size when compressing %d bytes" % (size),
            )
            self.assertEqual(compressed_bytes, b)

            # Now read little by little
            read_size = min(size // 4, 1024)
            uncompressed_stream = io.BytesIO(original_bytes)
            compressed_stream = utils.ZlibCompressedStream(uncompressed_stream)
            b = b""
            for c in iter(
                functools.partial(compressed_stream.read, read_size), b""
            ):
                b += c
            self.assertEqual(
                len(compressed_bytes),
                len(b),
                "Incorrect size when compressing %d bytes" % (size),
            )
            self.assertEqual(compressed_bytes, b)

    def test_zlib_streams_combined_randombytes(self):
        self._test_zlib_streams_combined(os.urandom)

    def test_zlib_streams_combined_zerobytes(self):
        self._test_zlib_streams_combined(lambda n: b"0" * n)

    def _test_zlib_streams_combined(self, gen_bytes):
        sizes = [2**x + y for x in range(1, 18) for y in (-1, 0, 1)]
        for size in sizes:
            original_bytes = gen_bytes(size)

            # Read the whole thing
            original_stream = io.BytesIO(original_bytes)
            compressed_stream = utils.ZlibCompressedStream(original_stream)
            uncompressed_stream = utils.ZlibUncompressedStream(
                compressed_stream
            )
            b = uncompressed_stream.read()
            self.assertEqual(size, len(b))
            self.assertEqual(original_bytes, b)

            # There's nothing left now to read
            for x in range(10):
                self.assertEqual(0, len(uncompressed_stream.read(100)))

            # Read with given number of bytes
            for n in (1, len(original_bytes) // 2, len(original_bytes)):
                these_bytes = original_bytes[0:n]

                original_stream = io.BytesIO(these_bytes)
                compressed_stream = utils.ZlibCompressedStream(original_stream)
                uncompressed_stream = utils.ZlibUncompressedStream(
                    compressed_stream
                )

                b = uncompressed_stream.read(n)
                self.assertEqual(n, len(b))
                self.assertEqual(these_bytes, b)

                # There's nothing left now to read
                for x in range(10):
                    self.assertEqual(0, len(uncompressed_stream.read(100)))

    def test_json_stream_simple_sequence(self):
        for s in ([0], [{}], ["a"], [{"oid": "A", "categoryType": "Data"}]):
            stream = utils.JSONStream(s)
            self.assertEqual(s, json.loads(stream.read(100).decode("utf8")))

    def test_json_stream_sequences(self):
        ref = [1, 2, 3]
        objects_list = [1, 2, 3]
        objects_tuple = (1, 2, 3)

        def objects_gen():
            yield 1
            yield 2
            yield 3

        for objects in (objects_list, objects_tuple, objects_gen()):
            stream = utils.JSONStream(objects)
            self.assertSequenceEqual(
                ref, json.loads(stream.read(100).decode("latin1"))
            )
            self.assertEqual(0, len(stream.read(100).decode("latin1")))

    def test_json_stream_simpleobj(self):
        sessionId = "some_id"
        for obj in (1, {"a": 2}, "b", {"sessionId": sessionId}):
            stream = utils.JSONStream(obj)
            self.assertEqual(
                obj, json.loads(stream.read(100).decode("latin1"))
            )
            self.assertEqual(0, len(stream.read(100).decode("latin1")))

    def test_get_dlg_root(self):
        # It should obey the DLG_ROOT environment variable
        old = os.environ.get("DLG_ROOT", None)
        os.environ["DLG_ROOT"] = tempfile.mkdtemp()
        self.assertEqual(utils.getDlgDir(), os.environ["DLG_ROOT"])
        os.rmdir(os.environ["DLG_ROOT"])
        if old:
            os.environ["DLG_ROOT"] = old
        else:
            del os.environ["DLG_ROOT"]



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2019
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import os
import tempfile
import unittest

from dlg.deploy import common
from dlg.manager import constants
from dlg.manager.session import SessionStates


from dlg.testutils import ManagerStarter

default_repro = {
    "rmode": "1",
    "RERUN": {
        "lg_blockhash": "x",
        "pgt_blockhash": "y",
        "pg_blockhash": "z",
    },
}
default_graph_repro = {
    "rmode": "1",
    "meta_data": {"repro_protocol": 0.1, "hashing_alg": "_sha3.sha3_256"},
    "merkleroot": "a",
    "RERUN": {
        "signature": "b",
    },
}


def add_test_reprodata(graph: list):
    for drop in graph:
        drop["reprodata"] = default_repro.copy()
    graph.append(default_graph_repro.copy())
    return graph


class CommonTestsBase(ManagerStarter):
    def _submit(self):
        pg = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "dlg.apps.simple.SleepApp",
                "inputs": ["A"],
                "outputs": ["C"],
            },
            {
                "oid": "C",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
        ]
        pg = add_test_reprodata(pg)
        for drop in pg:
            drop["node"] = "localhost"
            drop["island"] = "localhost"
        return common.submit(pg, "localhost", self.port)

    def assert_sessions_finished(self, status, *session_ids):
        for session_id in session_ids:
            self.assert_session_finished(status[session_id])

    def test_submit(self):
        self._submit()

    def test_monitor(self):
        session_id = self._submit()
        status = common.monitor_sessions(
            session_id, port=self.port, poll_interval=0.1
        )
        self.assert_session_finished(status)

    def test_monitor_all(self):
        session_id1 = self._submit()
        session_id2 = self._submit()
        status = common.monitor_sessions(port=self.port, poll_interval=0.1)
        self.assert_sessions_finished(status, session_id1, session_id2)

    def test_monitor_with_dumping(self):
        dump_path = tempfile.mktemp()
        session_id = self._submit()
        status = common.monitor_sessions(
            session_id=session_id,
            poll_interval=0.1,
            port=self.port,
            status_dump_path=dump_path,
        )
        self.assert_session_finished(status)
        self.assertTrue(os.path.exists(dump_path))
        os.remove(dump_path)


class TestDeployCommonNM(CommonTestsBase, unittest.TestCase):
    port = constants.NODE_DEFAULT_REST_PORT

    def assert_session_finished(self, status):
        self.assertEqual(SessionStates.FINISHED, status)

    def setUp(self):
        super(TestDeployCommonNM, self).setUp()
        self.nm_info = self.start_nm_in_thread()

    def tearDown(self):
        self.nm_info.stop()
        super(TestDeployCommonNM, self).tearDown()


class TestDeployCommonDIM(TestDeployCommonNM, unittest.TestCase):
    port = constants.ISLAND_DEFAULT_REST_PORT

    def assert_session_finished(self, status):
        self.assertTrue(all([s == SessionStates.FINISHED for s in status]))

    def setUp(self):
        super(TestDeployCommonDIM, self).setUp()
        self.dim_info = self.start_dim_in_thread()

    def tearDown(self):
        self.dim_info.stop()
        super(TestDeployCommonDIM, self).tearDown()


class TestDeployCommonMM(TestDeployCommonDIM, unittest.TestCase):
    port = constants.MASTER_DEFAULT_REST_PORT

    def setUp(self):
        super(TestDeployCommonMM, self).setUp()
        self.mm_info = self.start_mm_in_thread()

    def tearDown(self):
        self.mm_info.stop()
        super(TestDeployCommonMM, self).tearDown()



"""
This is an example of splitting a measurementset using the CASA MSTRANSFORM task
We need to develop a template based on it so it can get multiple spws based on
frequency ranges and widths
To run it standalone, change the directories, which are now hardcoded
"""

import datetime
import os
import time

if __name__ == "__main__":

    dt = datetime.datetime.now()
    timestr = dt.strftime("%Y-%m-%dT%H-%M-%S")

    ms_dir = "/scratch/partner1024/chiles/final_products/20131122_941_6_FINAL_PRODUCTS/13B-266.sb27261805.eb28549602.56618.334173599535_calibrated_deepfield.ms"
    # ms_dir = '/scratch/jason/13B-266.sb28624226.eb28625769.56669.43262586805_calibrated_deepfield_adios.ms'
    log_dir = "/home/cwu/chiles/processing"
    # log_dir = '/scratch/jason/tmp'
    output_vis = "/scratch/partner1024/chiles/split_vis/{0}".format(timestr)
    # output_vis = '/scratch/jason/tmp/split_vis/{0}'.format(timestr)
    log_file = "{0}/{1}_split_time.log".format(log_dir, timestr)

    os.mkdir(output_vis)
    gap = 4

    st = time.time()
    for i in range(1):
        mstransform(
            vis=ms_dir,
            outputvis="{0}/vis_{1}~{2}".format(
                output_vis, 1020 + i * gap, 1024 + i * gap
            ),
            start="{0}MHz".format(1020 + i * gap),
            width="15.625kHz",
            spw="2~2",
            nchan=256,
            regridms=T,
            restfreq="1420.405752MHz",
            mode="frequency",
            outframe="lsrk",
            interpolation="linear",
            veltype="radio",
            combinespws=True,
            nspw=1,
            createmms=False,
            datacolumn="data",
        )

    tt = time.time() - st
    f = open(log_file, "w")
    f.write("total split time = {0} seconds".format(tt))
    f.close()

    """
    mstransform(vis=/scratch/partner1024/chiles/final_products//20131025_951_4_FINAL_PRODUCTS/13B-266.sb27248272.eb28094627.56590.41308579861_calibrated_deepfield.ms,
    outputvis=/scratch/partner1024/chiles/split_vis/12345/20131025_951_4/vis_1024~1028,
    start=1024MHz,width=15.625,spw=2~2,nchan=256)
    
    mstransform(vis=/scratch/partner1024/chiles/final_products//20131025_951_4_FINAL_PRODUCTS/13B-266.sb27248272.eb28094627.56590.41308579861_calibrated_deepfield.ms,
    outputvis=/scratch/partner1024/chiles/split_vis/12345/20131025_951_4/vis_1028~1032,
    start=1028MHz,width=15.625,spw=2~2,nchan=256)
    
    mstransform(vis=/scratch/partner1024/chiles/final_products//20131025_951_4_FINAL_PRODUCTS/13B-266.sb27248272.eb28094627.56590.41308579861_calibrated_deepfield.ms,
    outputvis=/scratch/partner1024/chiles/split_vis/12345/20131025_951_4/vis_1032~1036,
    start=1032MHz,width=15.625,spw=2~2,nchan=256)
    """



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Contains a module translating physical graphs to kubernetes helm charts.
"""
import json
import logging
import os
import pathlib
import shutil
import subprocess
import sys
import threading
import time

import dlg
import yaml
from dlg.common.version import version as dlg_version
from dlg.constants import NODE_DEFAULT_REST_PORT
from dlg.deploy.common import submit
from dlg.deploy.deployment_utils import (
    find_node_ips,
    find_service_ips,
    find_pod_ips,
    wait_for_pods,
)
from dlg.dropmake import pg_generator
from dlg.restutils import RestClient
from dlg.common.k8s_utils import check_k8s_env

logger = logging.getLogger(__name__)


def _num_deployments_required(islands, nodes):
    machines = set()
    machines.update(islands)
    machines.update(nodes)
    return len(machines)


def _write_chart(
    chart_dir,
    name: str,
    chart_name: str,
    version: str,
    app_version: str,
    home: str,
    description,
    keywords: list,
    sources: list,
    kubeVersion: str,
):
    chart_info = {
        "apiVersion": "v2",
        "name": chart_name,
        "categoryType": "Application",
        "version": version,
        "appVersion": app_version,
        "home": home,
        "description": description,
        "keywords": keywords,
        "sources": sources,
        "kubeVersion": kubeVersion,
    }
    # TODO: Fix app_version quotations.
    with open(
        f"{chart_dir}{os.sep}{name}", "w", encoding="utf-8"
    ) as chart_file:
        yaml.dump(chart_info, chart_file)


def _write_values(chart_dir, config):
    with open(
        f"{chart_dir}{os.sep}custom-values.yaml", "w+", encoding="utf-8"
    ) as value_file:
        yaml.dump(config, value_file)
    logger.info("Written custom-values file.")


def _read_values(chart_dir):
    with open(
        f"{chart_dir}{os.sep}values.yaml", "r", encoding="utf-8"
    ) as old_file:
        data = yaml.safe_load(old_file)
    with open(
        f"{chart_dir}{os.sep}values.yaml", "r", encoding="utf-8"
    ) as custom_file:
        new_data = yaml.safe_load(custom_file)
    data.update(new_data)
    logger.info("Read yaml values file")
    return data


def _find_resources(pgt_data):
    pgt = json.loads(pgt_data)
    nodes = list(map(lambda x: x["node"], pgt))
    islands = list(map(lambda x: x["island"], pgt))
    islands = list(set(islands))
    nodes = list(set(nodes))
    return islands, nodes


class HelmClient:
    """
    Writes necessary files to launch job with kubernetes.
    """

    def __init__(
        self,
        deploy_name,
        chart_name="daliuge-daemon",
        deploy_dir="./",
        submit=True,
        chart_version="0.1.0",
        value_config=None,
        physical_graph_file=None,
        chart_vars=None,
    ):
        if value_config is None:
            value_config = dict()
        self._k8s_access = check_k8s_env()
        self._chart_name = chart_name
        self._chart_vars = {
            "name": "daliuge-daemon",
            "appVersion": "v1.0.0",
            "home": "https://github.com/ICRAR/daliuge/daliuge-k8s",
            "description": "DALiuGE k8s deployment",
            "keywords": ["daliuge", "workflow"],
            "sources": ["https://github.com/ICRAR/daliuge/daliuge-k8s"],
            "kubeVersion": ">=1.10.0-0",
        }
        if chart_vars is not None:
            self._chart_vars.update(chart_vars)
        self._deploy_dir = deploy_dir
        self._chart_dir = os.path.join(self._deploy_dir, "daliuge-daemon")
        self._chart_version = chart_version
        self._deploy_name = deploy_name
        self._submit = submit
        self._value_data = value_config if value_config is not None else {}
        self._submission_endpoint = None
        self._k8s_nodes = find_node_ips()
        self._num_machines = 1
        self._pod_details = {}
        if physical_graph_file is not None:
            self._set_physical_graph(physical_graph_file)

        # Copy in template files.
        library_root = pathlib.Path(
            os.path.dirname(dlg.__file__)
        ).parent.parent
        logger.debug(f"Helm chart copied to: {library_root}")
        if sys.version_info >= (3, 8):
            shutil.copytree(
                os.path.join(library_root, "daliuge-k8s", "helm"),
                self._deploy_dir,
                dirs_exist_ok=True,
            )
        else:
            shutil.copytree(
                os.path.join(library_root, "daliuge-k8s", "helm"),
                self._deploy_dir,
            )

    def _set_physical_graph(self, physical_graph_content, co_host=True):
        self._physical_graph_file = physical_graph_content
        self._islands, self._nodes = _find_resources(self._physical_graph_file)
        self._num_machines = _num_deployments_required(
            self._islands, self._nodes
        ) - (1 if co_host else 0)

    def _find_pod_details(self):
        # NOTE: +1 for the master.
        service_ips = find_service_ips(self._num_machines + 1)
        pod_ips = find_pod_ips(self._num_machines + 1)
        labels = sorted([str(x) for x in range(self._num_machines)])
        for i in range(len(labels)):
            self._pod_details[labels[i]] = {
                "ip": pod_ips[i],
                "svc": service_ips[i],
            }
        self._pod_details["master"] = {
            "ip": pod_ips[-1],
            "svc": service_ips[-1],
        }
        logger.debug(f"Pod details: {self._pod_details}")

    def create_helm_chart(self, physical_graph_content, co_host=True):
        """
        Translates a physical graph to a kubernetes helm chart.
        For now, it will just try to run everything in a single container.
        """
        if not self._k8s_access:
            raise RuntimeError("Cannot access k8s")
        # Add charts
        self._set_physical_graph(physical_graph_content, co_host)
        _write_chart(
            self._chart_dir,
            "Chart.yaml",
            self._chart_name,
            self._chart_version,
            dlg_version,
            self._chart_vars["home"],
            self._chart_vars["description"],
            self._chart_vars["keywords"],
            self._chart_vars["sources"],
            self._chart_vars["kubeVersion"],
        )
        # Update values.yaml
        _write_values(self._chart_dir, self._value_data)
        self._value_data = _read_values(self._chart_dir)
        # Update template

    def start_manager(self, manager_node):
        if not self._k8s_access:
            raise RuntimeError("Cannot access k8s")
        self._submission_endpoint = self._pod_details[manager_node]["svc"]
        client = RestClient(
            self._submission_endpoint,
            self._value_data["service"]["daemon"]["port"],
            timeout=30,
        )
        node_ips = [x["ip"] for x in self._pod_details.values()]
        print(node_ips)
        data = json.dumps({"nodes": node_ips}).encode("utf-8")
        time.sleep(5)
        logger.debug(f"Starting manager on {self._submission_endpoint}")
        client._POST(
            "/managers/island/start",
            content=data,
            content_type="application/json",
        ).read()
        client._POST(
            "/managers/master/start",
            content=data,
            content_type="application/json",
        ).read()

    def start_nodes(self):
        if not self._k8s_access:
            raise RuntimeError("Cannot access k8s")
        ips = [x["svc"] for x in self._pod_details.values()]
        ips.remove(self._pod_details["master"]["svc"])
        for ip in ips:
            client = RestClient(
                ip, self._value_data["service"]["daemon"]["port"], timeout=30
            )
            time.sleep(5)
            logger.debug(f"Starting node on {ip}")
            # node_ips = ['localhost'] + [x['ip'] for x in self._pod_details.values()]
            node_ips = [x["ip"] for x in self._pod_details.values()]
            # data = json.dumps({'nodes': ['localhost']}).encode('utf-8')
            data = json.dumps({"nodes": node_ips}).encode("utf-8")
            client._POST(
                "/managers/master/start",
                content=data,
                content_type="application/json",
            ).read()

    def launch_helm(self, co_host=False):
        """
        Launches the built helm chart using the most straightforward commands possible.
        Assumes all files are prepared and validated.
        """
        if not self._k8s_access:
            raise RuntimeError("Cannot access k8s")
        if self._submit:
            os.chdir(self._deploy_dir)
            _write_values(
                self._chart_dir,
                {"deploy_id": "master", "name": f"{self._chart_name}-master"},
            )
            instruction = (
                f"helm install {self._deploy_name}-master {self._chart_name}/  "
                f"--values {self._chart_name}{os.sep}custom-values.yaml"
            )
            process_return_string = subprocess.check_output(
                [instruction], shell=True
            ).decode("utf-8")
            logger.info(f"{process_return_string}")
            for i in range(self._num_machines):
                _write_values(
                    self._chart_dir,
                    {"deploy_id": i, "name": f"{self._chart_name}-{i}"},
                )
                instruction = (
                    f"helm install {self._deploy_name}-{i} {self._chart_name}/  "
                    f"--values {self._chart_name}{os.sep}custom-values.yaml"
                )
                process_return_string = subprocess.check_output(
                    [instruction], shell=True
                ).decode("utf-8")
                logger.info(f"{process_return_string}")
                # TODO: Check running nodes before launching another
            self._find_pod_details()
            if wait_for_pods(self._num_machines):
                self.start_manager("master")
                # self.start_nodes()
            else:
                logger.error("K8s pods did not start in timeframe allocated")
                self.teardown()
                raise RuntimeWarning(
                    "K8s pods did not start in timeframe allocated"
                )
        else:
            logger.info(
                f"Created helm chart {self._chart_name} in {self._deploy_dir}"
            )

    def teardown(self):
        if not self._k8s_access:
            raise RuntimeError("Cannot access k8s")
        for i in range(self._num_machines - 1, -1, -1):
            subprocess.check_output(
                [f"helm uninstall daliuge-daemon-{i}"], shell=True
            )
        subprocess.check_output(
            [f"helm uninstall daliuge-daemon-master"], shell=True
        )

    def _monitor(self, session_id=None):
        def _task():
            while True:
                try:
                    dlg.deploy.common.monitor_sessions(
                        session_id=session_id,
                        host=self._submission_endpoint,
                        port=NODE_DEFAULT_REST_PORT,
                    )
                    break
                except:
                    logger.exception(
                        "Monitoring failed, attempting to restart"
                    )

        threads = threading.Thread(target=_task)
        threads.start()
        return threads

    def submit_pgt(self):
        """
        There is a semi-dynamic element to fetching the IPs of Node(s) to deploy to.
        Hence, launching the chart and initiating graph execution have been de-coupled.
        """
        if not self._k8s_access:
            raise RuntimeError("Cannot access k8s")
        # TODO: Check all nodes are operational first.
        pgt_data = json.loads(self._physical_graph_file)
        node_ips = [x["ip"] for x in self._pod_details.values()]
        node_ips.remove(self._pod_details["master"]["ip"])
        node_ips = [self._pod_details["master"]["ip"]] + node_ips
        # node_ips = ['localhost']
        physical_graph = pg_generator.resource_map(
            pgt_data, node_ips, co_host_dim=True
        )
        # TODO: Add dumping to log-dir
        submit(
            physical_graph,
            self._submission_endpoint,
            port=NODE_DEFAULT_REST_PORT,
            skip_deploy=False,
        )

    def submit_and_monitor_pgt(self):
        """
        Combines submission and monitoring steps of a pgt.
        """
        session_id = self.submit_pgt()
        monitoring_thread = self._monitor(session_id)
        monitoring_thread.join()

    def submit_pg(self):
        """
        There is a semi-dynamic element to fetching the IPs of Node(s) to deploy to.
        Hence, launching the chart and initiating graph execution have been de-coupled.
        """
        if not self._k8s_access:
            raise RuntimeError("Cannot access k8s")
        # TODO: Check all nodes are operational first.
        pg_data = json.loads(self._physical_graph_file)
        # TODO: Add dumping to log-dir
        submit(
            pg_data,
            self._submission_endpoint,
            port=NODE_DEFAULT_REST_PORT,
            skip_deploy=False,
        )

    def submit_and_monitor_pg(self):
        """
        Combines submission and monitoring steps of a pg.
        """
        session_id = self.submit_pg()
        monitoring_thread = self._monitor(session_id)
        monitoring_thread.join()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import functools
import io
import os
import time
import unittest

from dlg import droputils
from dlg.apps.dynlib import DynlibApp, DynlibStreamApp, DynlibProcApp
from dlg.ddap_protocol import DROPRel, DROPLinkType, DROPStates
from dlg.data.drops.data_base import NullDROP
from dlg.data.drops.memory import InMemoryDROP

from .setp_up import build_shared_library
from ..manager import test_dm

_libname = "dynlib_example"
_libfname = "libdynlib_example.so"
_libpath = os.path.join(os.path.dirname(__file__), _libfname)
print_stats = 0
bufsize = 20 * 1024 * 1024


@unittest.skipUnless(
    build_shared_library(_libname, _libpath),
    "Example dynamic library not available",
)
class DynlibAppTest(unittest.TestCase):
    def test_simple_batch_copy(self):
        self._test_simple_copy(False)

    def test_simple_stream_copy(self):
        self._test_simple_copy(True)

    def _test_simple_copy(self, streaming):
        """
        Checks that the following graph works, both in streaming and batch mode:

        A ----> B ----> C
           |       +--> D
           |
           +--> E ----> F
                   +--> G

        Both B and E use the same dynamically loaded library and work with their
        input A to copy the inputs into their outputs.
        """

        # Build the graph
        a = (NullDROP if streaming else InMemoryDROP)("a", "a")
        b, e = (
            (DynlibStreamApp if streaming else DynlibApp)(
                x, x, lib=_libpath, print_stats=print_stats, bufsize=bufsize
            )
            for x in ("b", "e")
        )
        c, d, f, g = (InMemoryDROP(x, x) for x in ("c", "d", "f", "g"))
        for app, outputs in (b, (c, d)), (e, (f, g)):
            (app.addStreamingInput if streaming else app.addInput)(a)
            for o in outputs:
                app.addOutput(o)

        # ~100 MBs of data should be copied over from a to c and d via b, etc
        data = os.urandom(1024 * 1024) * 100
        reader = io.BytesIO(data)
        with droputils.DROPWaiterCtx(self, (c, d, f, g), 10):
            if streaming:
                # Write the data in chunks so we actually exercise multiple calls
                # to the data_written library call
                for datum in iter(
                    functools.partial(reader.read, 1024 * 1024), b""
                ):
                    a.write(datum)
            else:
                a.write(data)
            a.setCompleted()

        for drop in (c, d, f, g):
            drop_data = droputils.allDropContents(drop)
            self.assertEqual(
                len(data),
                len(drop_data),
                "Data from %r is not what we wanted :(" % (drop,),
            )
            self.assertEqual(data, drop_data)

    def test_cancel_dynlibprocapp(self):
        """Checks that we can cancel a long-running dynlib proc app"""

        a = DynlibProcApp("a", "a", lib=_libpath, sleep_seconds=10)
        # we don't really end up using it, so it can be anything
        dummy_rpc_endpoint = (None, None)
        a._rpc_endpoint = dummy_rpc_endpoint
        with droputils.DROPWaiterCtx(self, (), timeout=0):
            a.async_execute()

        time.sleep(1)
        t0 = time.time()
        a.cancel()
        self.assertLess(
            time.time() - t0,
            1,
            "Cancelled dynlibprocapp in less than a second",
        )
        self.assertEqual(DROPStates.CANCELLED, a.status)


class IntraNMMixIng(test_dm.NMTestsMixIn):
    # Indicate which particular application should the test use
    app = None

    def test_input_in_remote_nm(self):
        """
        A test similar in spirit to TestDM.test_runGraphOneDOPerDom, but where
        application B is a DynlibApp. This makes sure that DynlibApps work fine
        across Node Managers.

        NM #1      NM #2
        =======    =============
        | A --|----|-> B --> C |
        =======    =============
        """
        g1 = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            }
        ]
        g2 = [
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": self.app,
                "lib": _libpath,
                "print_stats": print_stats,
                "bufsize": bufsize,
            },
            {
                "oid": "C",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "producers": ["B"],
            },
        ]
        rels = [DROPRel("A", DROPLinkType.INPUT, "B")]
        a_data = os.urandom(32)
        self._test_runGraphInTwoNMs(g1, g2, rels, a_data, a_data)

    def test_output_in_remote_nm(self):
        """
        Like the above, but with this graph. In this case the output (instead of
        the input) is in a remote Node Manager.

        NM #1            NM #2
        =============    =======
        | A --> B --|----|-> C |
        =============    =======
        """
        g1 = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "consumers": ["B"],
            },
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": self.app,
                "lib": _libpath,
                "print_stats": print_stats,
                "bufsize": bufsize,
            },
        ]
        g2 = [
            {
                "oid": "C",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            }
        ]
        rels = [DROPRel("B", DROPLinkType.PRODUCER, "C")]
        a_data = os.urandom(32)
        self._test_runGraphInTwoNMs(g1, g2, rels, a_data, a_data)

    def test_multiple_inputs_in_remote_nm(self):
        """Like the above, but with this graph. In this case two inputs are
        located in a remote Node Manager.

        NM #1      NM #2
        =======    ===============
        | A --|----|-|           |
        |     |    | |-> C --> D |
        | B --|----|-|           |
        =======    ===============
        """
        g1 = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
            {
                "oid": "B",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
        ]
        g2 = [
            {
                "oid": "C",
                "categoryType": "Application",
                "dropclass": self.app,
                "lib": _libpath,
                "print_stats": print_stats,
                "bufsize": bufsize,
            },
            {
                "oid": "D",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "producers": ["C"],
            },
        ]
        rels = [
            DROPRel("A", DROPLinkType.INPUT, "C"),
            DROPRel("B", DROPLinkType.INPUT, "C"),
        ]
        a_data = os.urandom(32)
        self._test_runGraphInTwoNMs(
            g1,
            g2,
            rels,
            a_data,
            a_data * 2,
            root_oids=("A", "B"),
            leaf_oid="D",
        )


@unittest.skipUnless(
    build_shared_library(_libname, _libpath),
    "Example dynamic library not available",
)
class IntraNMDynlibAppTest(IntraNMMixIng, unittest.TestCase):
    app = "dlg.apps.dynlib.DynlibApp"


@unittest.skipUnless(
    build_shared_library(_libname, _libpath),
    "Example dynamic library not available",
)
class IntraNMDynlibProcAppTest(IntraNMMixIng, unittest.TestCase):
    app = "dlg.apps.dynlib.DynlibProcApp"

    def test_crashing_dynlib(self):
        """Like test_multiple_inputs_in_remote_nm, but C crashes"""
        g1 = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
            {
                "oid": "B",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
        ]
        g2 = [
            {
                "oid": "C",
                "categoryType": "Application",
                "dropclass": self.app,
                "lib": _libpath,
                "print_stats": print_stats,
                "bufsize": bufsize,
                "crash_and_burn": 1,
            },
            {
                "oid": "D",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "producers": ["C"],
            },
        ]
        rels = [
            DROPRel("A", DROPLinkType.INPUT, "C"),
            DROPRel("B", DROPLinkType.INPUT, "C"),
        ]
        a_data = os.urandom(32)
        self._test_runGraphInTwoNMs(
            g1,
            g2,
            rels,
            a_data,
            None,
            root_oids=("A", "B"),
            leaf_oid="D",
            expected_failures=("C", "D"),
        )
